# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from apptools._testing.optional_dependencies import (
    numpy as np,
    tables,
    requires_numpy,
    requires_tables,
)

if np is not None and tables is not None:
    from ..dict_node import H5DictNode
    from .utils import open_h5file, temp_h5_file, temp_file


NODE = "/dict_node"


@requires_tables
@requires_numpy
class DictNodeTestCase(unittest.TestCase):
    def test_create(self):
        with temp_h5_file() as h5:
            h5dict = H5DictNode.add_to_h5file(h5, NODE)
            h5dict["a"] = 1
            h5dict["b"] = 2

            assert h5dict["a"] == 1
            assert h5dict["b"] == 2

    def test_is_dict_node(self):
        with temp_h5_file() as h5:
            node = h5.create_dict(NODE, {})
            assert H5DictNode.is_dict_node(node._h5_group)

    def test_is_not_dict_node(self):
        with temp_h5_file() as h5:
            node = h5.create_group(NODE)
            assert not H5DictNode.is_dict_node(node)
            assert not H5DictNode.is_dict_node(node._h5_group)

    def test_create_with_data(self):
        with temp_h5_file() as h5:
            data = {"a": 10}

            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            assert h5dict["a"] == 10

    def test_create_with_array_data(self):
        foo = np.arange(100)
        bar = np.arange(150)

        with temp_h5_file() as h5:
            data = {"a": 10, "foo": foo, "bar": bar}

            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            assert h5dict["a"] == 10
            np.testing.assert_allclose(h5dict["foo"], foo)
            np.testing.assert_allclose(h5dict["bar"], bar)

    def test_load_saved_dict_node(self):
        with temp_file() as filename:
            # Write data to new dict node and close.
            with open_h5file(filename, "w") as h5:
                h5dict = H5DictNode.add_to_h5file(h5, NODE)
                h5dict["a"] = 1

            # Read dict node and make sure the data was saved.
            with open_h5file(filename, mode="r+") as h5:
                h5dict = h5[NODE]
                assert h5dict["a"] == 1
                # Change data for next test
                h5dict["a"] = 2

            # Check that data is modified by the previous write.
            with open_h5file(filename) as h5:
                h5dict = h5[NODE]
                assert h5dict["a"] == 2

    def test_load_saved_dict_node_with_array(self):
        arr = np.arange(100)
        arr1 = np.arange(200)

        with temp_file() as filename:
            # Write data to new dict node and close.
            with open_h5file(filename, "w") as h5:
                h5dict = H5DictNode.add_to_h5file(h5, NODE)
                h5dict["arr"] = arr

            # Read dict node and make sure the data was saved.
            with open_h5file(filename, mode="r+") as h5:
                h5dict = h5[NODE]
                np.testing.assert_allclose(h5dict["arr"], arr)
                # Change data for next test
                h5dict["arr"] = arr1
                h5dict["arr_old"] = arr

            # Check that data is modified by the previous write.
            with open_h5file(filename) as h5:
                h5dict = h5[NODE]
                np.testing.assert_allclose(h5dict["arr"], arr1)
                np.testing.assert_allclose(h5dict["arr_old"], arr)
                # Make sure that arrays come back as arrays
                assert isinstance(h5dict["arr"], np.ndarray)
                assert isinstance(h5dict["arr_old"], np.ndarray)

    def test_keys(self):
        with temp_h5_file() as h5:
            keys = set(("hello", "world", "baz1"))
            data = dict((n, 1) for n in keys)
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            assert set(h5dict.keys()) == keys

    def test_delete_item(self):
        with temp_h5_file() as h5:
            data = dict(a=10)
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            del h5dict["a"]
            assert "a" not in h5dict

    def test_delete_array(self):
        with temp_h5_file() as h5:
            data = dict(a=np.arange(10))
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            del h5dict["a"]
            assert "a" not in h5dict
            assert "a" not in h5[NODE]

    def test_auto_flush(self):
        with temp_h5_file() as h5:
            data = dict(a=1)
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            # Overwrite existing data, which should get written to disk.
            new_data = dict(b=2)
            h5dict.data = new_data
            # Load data from disk to check that data was automatically flushed.
            h5dict_from_disk = h5[NODE]
            assert "a" not in h5dict_from_disk
            assert h5dict_from_disk["b"] == 2

    def test_auto_flush_off(self):
        with temp_h5_file() as h5:
            data = dict(a=1)
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data, auto_flush=False)
            # Overwrite existing data, but don't write to disk.
            new_data = dict(b=2)
            h5dict.data = new_data
            # Load data from disk to check that it's unchanged.
            h5dict_from_disk = h5[NODE]
            assert h5dict_from_disk["a"] == 1
            assert "b" not in h5dict_from_disk
            # Manually flush, and check that data was written
            h5dict.flush()
            h5dict_from_disk = h5[NODE]
            assert "a" not in h5dict_from_disk
            assert h5dict_from_disk["b"] == 2

    def test_undefined_key(self):
        with temp_h5_file() as h5:
            data = dict(a="int")
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            with self.assertRaises(KeyError):
                del h5dict["b"]

    def test_basic_dtypes(self):
        with temp_h5_file() as h5:
            data = dict(a_int=1, a_float=1.0, a_str="abc")
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            assert isinstance(h5dict["a_int"], int)
            assert isinstance(h5dict["a_float"], float)
            assert isinstance(h5dict["a_str"], str)

    def test_mixed_type_list(self):
        with temp_h5_file() as h5:
            data = dict(a=[1, 1.0, "abc"])
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            for value, dtype in zip(
                h5dict["a"], (int, float, str)
            ):
                assert isinstance(value, dtype)

    def test_dict(self):
        with temp_h5_file() as h5:
            data = dict(a=dict(b=1, c=2))
            h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
            sub_dict = h5dict["a"]
            assert sub_dict["b"] == 1
            assert sub_dict["c"] == 2

    def test_wrap_self_raises_error(self):
        with temp_h5_file() as h5:
            H5DictNode.add_to_h5file(h5, NODE)
            node = h5[NODE]
            with self.assertRaises(AssertionError):
                H5DictNode(node)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" An instance editor that allows total control over widget creation. """


# Enthought library imports.
from traits.api import Any
from traitsui.api import EditorFactory
from traitsui.toolkit import toolkit_object
Editor = toolkit_object('editor:Editor')


class _WidgetEditor(Editor):
    """ An instance editor that allows total control over widget creation. """

    #### '_WidgetEditor' interface ############################################

    # The toolkit-specific parent of the editor.
    parent = Any

    ###########################################################################
    # '_WidgetEditor' interface.
    ###########################################################################

    def init(self, parent):
        """ Initialize the editor. """

        self.parent = parent

        # fixme: What if there are no pages?!?
        page = self.object.pages[0]

        # Create the editor's control.
        self.control = page.create_control(parent)

        # Listen for the page being changed.
        self.object.on_trait_change(self._on_page_changed, "selected_page")

    def dispose(self):
        """ Dispose of the editor. """

        page = self.object.selected_page
        page.destroy_control()

    def update_editor(self):
        """ Update the editor. """

        pass

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _on_page_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        if old is not None:
            old.destroy_control()

        if new is not None:
            self.control = new.create_control(self.parent)


class WidgetEditor(EditorFactory):
    """ A factory widget editors. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __call__(self, *args, **traits):
        """ Call the object. """

        return self.trait_set(**traits)

    ###########################################################################
    # 'EditorFactory' interface.
    ###########################################################################

    def simple_editor(self, ui, object, name, description, parent):
        """ Create a simple editor. """

        editor = _WidgetEditor(
            parent,
            factory=self,
            ui=ui,
            object=object,
            name=name,
            description=description,
        )

        return editor

    custom_editor = simple_editor
    text_editor = simple_editor
    readonly_editor = simple_editor



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from .i_preferences import IPreferences

from .package_globals import get_default_preferences, set_default_preferences
from .preferences import Preferences
from .preference_binding import PreferenceBinding, bind_preference
from .preferences_helper import PreferencesHelper
from .scoped_preferences import ScopedPreferences



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""
Globals for the scripting package.
"""


# The global recorder.
_recorder = None


def get_recorder():
    """Return the global recorder.  Does not create a new one if none
    exists.
    """
    global _recorder
    return _recorder


def set_recorder(rec):
    """Set the global recorder instance."""
    global _recorder
    _recorder = rec



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# -------------------------------------------------------------------------------
#  Imports:
# -------------------------------------------------------------------------------

import sys

from traits.api import Trait, TraitHandler, TraitFactory

from traits.trait_base import class_of, get_module_name

from apptools.naming.api import Binding


# -------------------------------------------------------------------------------
#  'NamingInstance' trait factory:
# -------------------------------------------------------------------------------


def NamingInstance(klass=None, value="", allow_none=False, **metadata):
    metadata.setdefault("copy", "deep")
    return Trait(
        value,
        NamingTraitHandler(
            klass, or_none=allow_none, module=get_module_name()
        ),
        **metadata
    )


NamingInstance = TraitFactory(NamingInstance)

# -------------------------------------------------------------------------------
#  'NamingTraitHandler' class:
# -------------------------------------------------------------------------------


class NamingTraitHandler(TraitHandler):

    # ---------------------------------------------------------------------------
    #  Initializes the object:
    # ---------------------------------------------------------------------------

    def __init__(self, aClass, or_none, module):
        """Initializes the object."""
        self.or_none = or_none is not False
        self.module = module
        self.aClass = aClass
        if (aClass is not None) and (
            not isinstance(aClass, (str, type))
        ):
            self.aClass = aClass.__class__

    def validate(self, object, name, value):
        if isinstance(value, str):
            if value == "":
                if self.or_none:
                    return ""
                else:
                    self.validate_failed(object, name, value)
            try:
                value = self._get_binding_for(value)
            except:  # noqa: E722
                self.validate_failed(object, name, value)

        if isinstance(self.aClass, str):
            self.resolve_class(object, name, value)

        if isinstance(value, Binding) and (
            (self.aClass is None) or isinstance(value.obj, self.aClass)
        ):
            return value.namespace_name
        self.validate_failed(object, name, value)

    def info(self):
        aClass = self.aClass
        if aClass is None:
            result = "path"
        else:
            if type(aClass) is not str:
                aClass = aClass.__name__
            result = "path to an instance of " + class_of(aClass)
        if self.or_none is None:
            return result + " or an empty string"
        return result

    def validate_failed(self, object, name, value):
        if not isinstance(value, type):
            msg = "class %s" % value.__class__.__name__
        else:
            msg = "%s (i.e. %s)" % (str(type(value))[1:-1], repr(value))
        self.error(object, name, msg)

    def get_editor(self, trait):
        if self.editor is None:
            from traitsui.api import DropEditor

            self.editor = DropEditor(
                klass=self.aClass, binding=True, readonly=False
            )
        return self.editor

    def post_setattr(self, object, name, value):
        other = None
        if value != "":
            other = self._get_binding_for(value).obj
        object.__dict__[name + "_"] = other

    def _get_binding_for(self, value):

        result = None

        # FIXME: The following code makes this whole component have a
        # dependency on envisage, and worse, assumes the use of a particular
        # project plugin!  This is horrible and should be refactored out,
        # possibly to a custom sub-class of whoever needs this behavior.
        try:
            from envisage import get_application

            workspace = get_application().service_registry.get_service(
                "envisage.project.IWorkspace"
            )
            result = workspace.lookup_binding(value)
        except ImportError:
            pass

        return result

    def resolve_class(self, object, name, value):
        aClass = self.find_class()
        if aClass is None:
            self.validate_failed(object, name, value)
        self.aClass = aClass

        # fixme: The following is quite ugly, because it wants to try and fix
        # the trait referencing this handler to use the 'fast path' now that
        # the actual class has been resolved. The problem is finding the trait,
        # especially in the case of List(Instance('foo')), where the
        # object.base_trait(...) value is the List trait, not the Instance
        # trait, so we need to check for this and pull out the List
        # 'item_trait'. Obviously this does not extend well to other traits
        # containing nested trait references (Dict?)...
        trait = object.base_trait(name)
        handler = trait.handler
        if (handler is not self) and hasattr(handler, "item_trait"):
            trait = handler.item_trait
        trait.validate(self.fast_validate)

    def find_class(self):
        module = self.module
        aClass = self.aClass
        col = aClass.rfind(".")
        if col >= 0:
            module = aClass[:col]
            aClass = aClass[col + 1:]
        theClass = getattr(sys.modules.get(module), aClass, None)
        if (theClass is None) and (col >= 0):
            try:
                theClass = getattr(__import__(module), aClass, None)
            except Exception:
                pass
        return theClass



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests for preferences nodes. """


# Standard library imports.
import os
import tempfile
import unittest
from os.path import join

# Major package imports.
from importlib_resources import files

# Enthought library imports.
from apptools.preferences.api import Preferences


# This module's package.
PKG = "apptools.preferences.tests"


class PreferencesTestCase(unittest.TestCase):
    """ Tests for preferences nodes. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.preferences = Preferences()

        # The filename of the example preferences file.
        self.example = os.fspath(files(PKG) / "example.ini")

        # A temporary directory that can safely be written to.
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        # Remove the temporary directory.
        os.rmdir(self.tmpdir)

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_package_global_default_preferences(self):
        """ package global default preferences """

        from apptools.preferences.api import get_default_preferences
        from apptools.preferences.api import set_default_preferences

        set_default_preferences(self.preferences)
        self.assertEqual(self.preferences, get_default_preferences())

    def test_get_and_set_str(self):
        """ get and set str """

        p = self.preferences

        # Set a string preference.
        p.set("acme.ui.bgcolor", "blue")
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))

    def test_get_and_set_int(self):
        """ get and set int """

        p = self.preferences

        # Note that we can pass an actual 'int' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set("acme.ui.width", 50)
        self.assertEqual("50", p.get("acme.ui.width"))

    def test_get_and_set_float(self):
        """ get and set float """

        p = self.preferences

        # Note that we can pass an actual 'flaot' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set("acme.ui.ratio", 1.0)
        self.assertEqual("1.0", p.get("acme.ui.ratio"))

    def test_get_and_set_bool(self):
        """ get and set bool """

        p = self.preferences

        # Note that we can pass an actual 'bool' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set("acme.ui.visible", True)
        self.assertEqual("True", p.get("acme.ui.visible"))

    def test_get_and_set_list_of_str(self):
        """ get and set list of str """

        p = self.preferences

        # Note that we can pass an actual 'int' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set("acme.ui.names", ["fred", "wilma", "barney"])
        self.assertEqual("['fred', 'wilma', 'barney']", p.get("acme.ui.names"))

    def test_get_and_set_list_of_int(self):
        """ get and set list of int """

        p = self.preferences

        # Note that we can pass an actual 'int' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set("acme.ui.offsets", [1, 2, 3])
        self.assertEqual("[1, 2, 3]", p.get("acme.ui.offsets"))

    def test_empty_path(self):
        """ empty path """

        p = self.preferences

        self.assertRaises(ValueError, p.get, "")
        self.assertRaises(ValueError, p.remove, "")
        self.assertRaises(ValueError, p.set, "", "a value")

    def test_default_values(self):
        """ default values """

        p = self.preferences

        # Try non-existent names to get the default-default!
        self.assertIsNone(p.get("bogus"))
        self.assertIsNone(p.get("acme.bogus"))
        self.assertIsNone(p.get("acme.ui.bogus"))

        # Try non-existent names to get the specified default.
        self.assertEqual("a value", p.get("bogus", "a value"))
        self.assertEqual("a value", p.get("acme.bogus", "a value"))
        self.assertEqual("a value", p.get("acme.ui.bogus", "a value"))

    def test_keys(self):
        """ keys """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], list(p.keys()))

        # Set some preferences in the node.
        p.set("a", "1")
        p.set("b", "2")
        p.set("c", "3")

        keys = sorted(p.keys())

        self.assertEqual(["a", "b", "c"], keys)

        # Set some preferences in a child node.
        p.set("acme.a", "1")
        p.set("acme.b", "2")
        p.set("acme.c", "3")

        keys = sorted(p.keys("acme"))

        self.assertEqual(["a", "b", "c"], keys)

        # And, just to be sure, in a child of the child node ;^)
        p.set("acme.ui.a", "1")
        p.set("acme.ui.b", "2")
        p.set("acme.ui.c", "3")

        keys = sorted(p.keys("acme.ui"))

        self.assertEqual(["a", "b", "c"], keys)

        # Test keys of a non-existent node.
        self.assertEqual([], p.keys("bogus"))
        self.assertEqual([], p.keys("bogus.blargle"))
        self.assertEqual([], p.keys("bogus.blargle.foogle"))

    def test_node(self):
        """ node """

        p = self.preferences

        # Try an empty path.
        self.assertEqual(p, p.node())

        # Try a simple path.
        node = p.node("acme")
        self.assertIsNotNone(node)
        self.assertEqual("acme", node.name)
        self.assertEqual("acme", node.path)
        self.assertEqual(p, node.parent)

        # Make sure we get the same node each time we ask for it!
        self.assertEqual(node, p.node("acme"))

        # Try a nested path.
        node = p.node("acme.ui")
        self.assertIsNotNone(node)
        self.assertEqual("ui", node.name)
        self.assertEqual("acme.ui", node.path)
        self.assertEqual(p.node("acme"), node.parent)

        # And just to be sure, a really nested path.
        node = p.node("acme.ui.splash_screen")
        self.assertIsNotNone(node)
        self.assertEqual("splash_screen", node.name)
        self.assertEqual("acme.ui.splash_screen", node.path)
        self.assertEqual(p.node("acme.ui"), node.parent)

    def test_node_exists(self):
        """ node exists """

        p = self.preferences

        self.assertTrue(p.node_exists())
        self.assertFalse(p.node_exists("acme"))

        p.node("acme")
        self.assertTrue(p.node_exists("acme"))

    def test_node_names(self):
        """ node names """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.node_names())

        # Add some nodes.
        p.node("a")
        p.node("b")
        p.node("c")

        names = sorted(p.node_names())

        self.assertEqual(["a", "b", "c"], names)

        # Creatd some nodes in a child node.
        p.node("acme.a")
        p.node("acme.b")
        p.node("acme.c")

        names = sorted(p.node_names("acme"))

        self.assertEqual(["a", "b", "c"], names)

        # And, just to be sure, in a child of the child node ;^)
        p.node("acme.ui.a")
        p.node("acme.ui.b")
        p.node("acme.ui.c")

        names = sorted(p.node_names("acme.ui"))

        self.assertEqual(["a", "b", "c"], names)

        # Test keys of a non-existent node.
        self.assertEqual([], p.node_names("bogus"))
        self.assertEqual([], p.node_names("bogus.blargle"))
        self.assertEqual([], p.node_names("bogus.blargle.foogle"))

    def test_clear(self):
        """ clear """

        p = self.preferences

        # Set some values.
        p.set("acme.ui.bgcolor", "blue")
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))
        p.set("acme.ui.width", 100)
        self.assertEqual("100", p.get("acme.ui.width"))

        # Clear all preferences from the node.
        p.clear("acme.ui")
        self.assertIsNone(p.get("acme.ui.bgcolor"))
        self.assertIsNone(p.get("acme.ui.width"))
        self.assertEqual(0, len(p.keys("acme.ui")))

    def test_remove(self):
        """ remove """

        p = self.preferences

        # Set a value.
        p.set("acme.ui.bgcolor", "blue")
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))

        # Remove it.
        p.remove("acme.ui.bgcolor")
        self.assertIsNone(p.get("acme.ui.bgcolor"))

        # Make sure we can't remove nodes!
        p.remove("acme.ui")
        self.assertTrue(p.node_exists("acme.ui"))

    def test_flush(self):
        """ flush """

        p = self.preferences

        # A temporary .ini file for this test.
        tmp = join(self.tmpdir, "tmp.ini")

        # This could be set in the constructor of course, its just here we
        # want to use the instance declared in 'setUp'.
        p.filename = tmp

        try:
            # Load the preferences from an 'ini' file.
            p.load(self.example)

            # Flush it.
            p.flush()

            # Load it into a new node.
            p = Preferences()
            p.load(tmp)

            # Make sure it was all loaded!
            self.assertEqual("blue", p.get("acme.ui.bgcolor"))
            self.assertEqual("50", p.get("acme.ui.width"))
            self.assertEqual("1.0", p.get("acme.ui.ratio"))
            self.assertEqual("True", p.get("acme.ui.visible"))
            self.assertEqual("acme ui", p.get("acme.ui.description"))
            self.assertEqual("[1, 2, 3, 4]", p.get("acme.ui.offsets"))
            self.assertEqual("['joe', 'fred', 'jane']", p.get("acme.ui.names"))
            self.assertEqual("splash", p.get("acme.ui.splash_screen.image"))
            self.assertEqual("red", p.get("acme.ui.splash_screen.fgcolor"))

        finally:
            # Clean up!
            os.remove(tmp)

    def test_load(self):
        """ load """

        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)

        # Make sure it was all loaded!
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))
        self.assertEqual("50", p.get("acme.ui.width"))
        self.assertEqual("1.0", p.get("acme.ui.ratio"))
        self.assertEqual("True", p.get("acme.ui.visible"))
        self.assertEqual("acme ui", p.get("acme.ui.description"))
        self.assertEqual("[1, 2, 3, 4]", p.get("acme.ui.offsets"))
        self.assertEqual("['joe', 'fred', 'jane']", p.get("acme.ui.names"))
        self.assertEqual("splash", p.get("acme.ui.splash_screen.image"))
        self.assertEqual("red", p.get("acme.ui.splash_screen.fgcolor"))

    def test_load_with_filename_trait_set(self):
        """ load with filename trait set """

        p = self.preferences
        p.filename = self.example

        # Load the preferences from an 'ini' file.
        p.load()

        # Make sure it was all loaded!
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))
        self.assertEqual("50", p.get("acme.ui.width"))
        self.assertEqual("1.0", p.get("acme.ui.ratio"))
        self.assertEqual("True", p.get("acme.ui.visible"))
        self.assertEqual("acme ui", p.get("acme.ui.description"))
        self.assertEqual("[1, 2, 3, 4]", p.get("acme.ui.offsets"))
        self.assertEqual("['joe', 'fred', 'jane']", p.get("acme.ui.names"))
        self.assertEqual("splash", p.get("acme.ui.splash_screen.image"))
        self.assertEqual("red", p.get("acme.ui.splash_screen.fgcolor"))

        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)

        # Make sure it was all loaded!
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))
        self.assertEqual("50", p.get("acme.ui.width"))
        self.assertEqual("1.0", p.get("acme.ui.ratio"))
        self.assertEqual("True", p.get("acme.ui.visible"))
        self.assertEqual("acme ui", p.get("acme.ui.description"))
        self.assertEqual("[1, 2, 3, 4]", p.get("acme.ui.offsets"))
        self.assertEqual("['joe', 'fred', 'jane']", p.get("acme.ui.names"))
        self.assertEqual("splash", p.get("acme.ui.splash_screen.image"))
        self.assertEqual("red", p.get("acme.ui.splash_screen.fgcolor"))

    def test_save(self):
        """ save """

        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)

        # Make sure it was all loaded!
        self.assertEqual("blue", p.get("acme.ui.bgcolor"))
        self.assertEqual("50", p.get("acme.ui.width"))
        self.assertEqual("1.0", p.get("acme.ui.ratio"))
        self.assertEqual("True", p.get("acme.ui.visible"))
        self.assertEqual("acme ui", p.get("acme.ui.description"))
        self.assertEqual("[1, 2, 3, 4]", p.get("acme.ui.offsets"))
        self.assertEqual("['joe', 'fred', 'jane']", p.get("acme.ui.names"))
        self.assertEqual("splash", p.get("acme.ui.splash_screen.image"))
        self.assertEqual("red", p.get("acme.ui.splash_screen.fgcolor"))

        # Make a change.
        p.set("acme.ui.bgcolor", "yellow")

        # Save it to another file.
        tmp = join(self.tmpdir, "tmp.ini")
        p.save(tmp)

        try:
            # Load it into a new node.
            p = Preferences()
            p.load(tmp)

            # Make sure it was all loaded!
            self.assertEqual("yellow", p.get("acme.ui.bgcolor"))
            self.assertEqual("50", p.get("acme.ui.width"))
            self.assertEqual("1.0", p.get("acme.ui.ratio"))
            self.assertEqual("True", p.get("acme.ui.visible"))
            self.assertEqual("acme ui", p.get("acme.ui.description"))
            self.assertEqual("[1, 2, 3, 4]", p.get("acme.ui.offsets"))
            self.assertEqual("['joe', 'fred', 'jane']", p.get("acme.ui.names"))
            self.assertEqual("splash", p.get("acme.ui.splash_screen.image"))
            self.assertEqual("red", p.get("acme.ui.splash_screen.fgcolor"))

        finally:
            # Clean up!
            os.remove(tmp)

    def SKIPtest_dump(self):
        """ dump """

        # This make look like a weird test, since we don't ever actually check
        # anything, but it is useful for people to see the structure of a
        # preferences hierarchy.
        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)
        p.dump()

    def test_get_inherited(self):
        """ get inherited """

        p = self.preferences

        # Set a string preference.
        p.set("bgcolor", "red")
        p.set("acme.bgcolor", "green")
        p.set("acme.ui.bgcolor", "blue")

        self.assertEqual("blue", p.get("acme.ui.bgcolor", inherit=True))

        # Now remove the 'lowest' layer.
        p.remove("acme.ui.bgcolor")
        self.assertEqual("green", p.get("acme.ui.bgcolor", inherit=True))

        # And the next one.
        p.remove("acme.bgcolor")
        self.assertEqual("red", p.get("acme.ui.bgcolor", inherit=True))

        # And the last one.
        p.remove("bgcolor")
        self.assertEqual(None, p.get("acme.ui.bgcolor", inherit=True))

    def test_add_listener(self):
        """ add listener """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key = key
            listener.old = old
            listener.new = new

        # Add a listener.
        p.add_preferences_listener(listener, "acme.ui")

        # Set a value and make sure the listener was called.
        p.set("acme.ui.bgcolor", "blue")
        self.assertEqual(p.node("acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertIsNone(listener.old)
        self.assertEqual("blue", listener.new)

        # Set it to another value to make sure we get the 'old' value
        # correctly.
        p.set("acme.ui.bgcolor", "red")
        self.assertEqual(p.node("acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertEqual("blue", listener.old)
        self.assertEqual("red", listener.new)

    def test_remove_listener(self):
        """ remove listener """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key = key
            listener.old = old
            listener.new = new

        # Add a listener.
        p.add_preferences_listener(listener, "acme.ui")

        # Set a value and make sure the listener was called.
        p.set("acme.ui.bgcolor", "blue")
        self.assertEqual(p.node("acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertIsNone(listener.old)
        self.assertEqual("blue", listener.new)

        # Remove the listener.
        p.remove_preferences_listener(listener, "acme.ui")

        # Set a value and make sure the listener was *not* called.
        listener.node = None

        p.set("acme.ui.bgcolor", "blue")
        self.assertIsNone(listener.node)

    def test_set_with_same_value(self):
        """ set with same value """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key = key
            listener.old = old
            listener.new = new

        # Add a listener.
        p.add_preferences_listener(listener, "acme.ui")

        # Set a value and make sure the listener was called.
        p.set("acme.ui.bgcolor", "blue")
        self.assertEqual(p.node("acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertIsNone(listener.old)
        self.assertEqual("blue", listener.new)

        # Clear out the listener.
        listener.node = None

        # Set the same value and make sure the listener *doesn't* get called.
        p.set("acme.ui.bgcolor", "blue")
        self.assertIsNone(listener.node)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Prints a stack trace every time it is called but does not halt execution
    of the application.

    Copied from Uche Ogbuji's blog
"""

# Standard library imports.
import inspect
from io import StringIO


def log_point(msg="\n"):
    stack = inspect.stack()
    # get rid of logPoint's part of the stack:
    stack = stack[1:]
    stack.reverse()
    output = StringIO()
    if msg:
        output.write(str(msg) + "\n")
    for stackLine in stack:
        frame, filename, line, funcname, lines, unknown = stackLine
        if filename.endswith("/unittest.py"):
            # unittest.py code is a boring part of the traceback
            continue
        if filename.startswith("./"):
            filename = filename[2:]
        output.write("%s:%s in %s:\n" % (filename, line, funcname))
        if lines:
            output.write("  %s\n" % "".join(lines)[:-1])
    s = output.getvalue()

    return s



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Package-scope globals.

The default preferences node is currently used by 'PreferencesHelper' and
'PreferencesBinding' instances if no specific preferences node is set. This
makes it easy for them to access the root node of an application-wide
preferences hierarchy.

"""


# The default preferences node.
_default_preferences = None


def get_default_preferences():
    """ Get the default preferences node. """

    return _default_preferences


def set_default_preferences(default_preferences):
    """ Set the default preferences node. """

    global _default_preferences

    _default_preferences = default_preferences

    # For convenience.
    return _default_preferences



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" The interface for pages in a preferences dialog. """


# Enthought library imports.
from traits.api import Interface, Str


class IPreferencesPage(Interface):
    """ The interface for pages in a preferences dialog. """

    # The page's category (e.g. 'General/Appearence'). The empty string means
    # that this is a top-level page.
    category = Str

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = Str

    # The page name (this is what is shown in the preferences dialog).
    name = Str

    def apply(self):
        """ Apply the page's preferences. """
        pass



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests file operations. """


# Standard library imports.
import os
import shutil
import stat
import unittest

# Enthought library imports.
from apptools.io import File


class FileTestCase(unittest.TestCase):
    """ Tests file operations on a local file system. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        shutil.rmtree("data", ignore_errors=True)

        os.mkdir("data")

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        shutil.rmtree("data")

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_properties(self):
        """ file properties """

        # Properties of a non-existent file.
        f = File("data/bogus.xx")

        self.assertIn(os.path.abspath(os.path.curdir), f.absolute_path)
        self.assertIsNone(f.children)
        self.assertEqual(f.ext, ".xx")
        self.assertFalse(f.exists)
        self.assertFalse(f.is_file)
        self.assertFalse(f.is_folder)
        self.assertFalse(f.is_package)
        self.assertFalse(f.is_readonly)
        self.assertEqual(f.mime_type, "content/unknown")
        self.assertEqual(f.name, "bogus")
        self.assertEqual(f.parent.path, "data")
        self.assertEqual(f.path, "data/bogus.xx")
        self.assertIn(os.path.abspath(os.path.curdir), f.url)
        self.assertEqual(str(f), "File(%s)" % f.path)

        # Properties of an existing file.
        f = File("data/foo.txt")
        f.create_file()

        self.assertIn(os.path.abspath(os.path.curdir), f.absolute_path)
        self.assertIsNone(f.children)
        self.assertEqual(f.ext, ".txt")
        self.assertTrue(f.exists)
        self.assertTrue(f.is_file)
        self.assertFalse(f.is_folder)
        self.assertFalse(f.is_package)
        self.assertFalse(f.is_readonly)
        self.assertEqual(f.mime_type, "text/plain")
        self.assertEqual(f.name, "foo")
        self.assertEqual(f.parent.path, "data")
        self.assertEqual(f.path, "data/foo.txt")
        self.assertIn(os.path.abspath(os.path.curdir), f.url)

        # Make it readonly.
        os.chmod(f.path, stat.S_IRUSR)
        self.assertTrue(f.is_readonly)

        # And then make it NOT readonly so that we can delete it at the end of
        # the test!
        os.chmod(f.path, stat.S_IRUSR | stat.S_IWUSR)
        self.assertFalse(f.is_readonly)

    def test_copy(self):
        """ file copy """

        content = 'print("Hello World!")\n'

        f = File("data/foo.txt")
        self.assertFalse(f.exists)

        # Create the file.
        f.create_file(content)
        self.assertTrue(f.exists)
        self.assertRaises(ValueError, f.create_file, content)

        self.assertIsNone(f.children)
        self.assertEqual(f.ext, ".txt")
        self.assertTrue(f.is_file)
        self.assertFalse(f.is_folder)
        self.assertEqual(f.mime_type, "text/plain")
        self.assertEqual(f.name, "foo")
        self.assertEqual(f.path, "data/foo.txt")

        # Copy the file.
        g = File("data/bar.txt")
        self.assertFalse(g.exists)

        f.copy(g)
        self.assertTrue(g.exists)

        self.assertIsNone(g.children)
        self.assertEqual(g.ext, ".txt")
        self.assertTrue(g.is_file)
        self.assertFalse(g.is_folder)
        self.assertEqual(g.mime_type, "text/plain")
        self.assertEqual(g.name, "bar")
        self.assertEqual(g.path, "data/bar.txt")

        # Attempt to copy a non-existent file (should do nothing).
        f = File("data/bogus.xx")
        self.assertFalse(f.exists)

        g = File("data/bogus_copy.txt")
        self.assertFalse(g.exists)

        f.copy(g)
        self.assertFalse(g.exists)

    def test_create_file(self):
        """ file creation """

        content = 'print("Hello World!")\n'

        f = File("data/foo.txt")
        self.assertFalse(f.exists)

        # Create the file.
        f.create_file(content)
        self.assertTrue(f.exists)
        with open(f.path) as file:
            self.assertEqual(file.read(), content)

        # Try to create it again.
        self.assertRaises(ValueError, f.create_file, content)

    def test_delete(self):
        """ file deletion """

        content = 'print("Hello World!")\n'

        f = File("data/foo.txt")
        self.assertFalse(f.exists)

        # Create the file.
        f.create_file(content)
        self.assertTrue(f.exists)
        self.assertRaises(ValueError, f.create_file, content)

        self.assertIsNone(f.children)
        self.assertEqual(f.ext, ".txt")
        self.assertTrue(f.is_file)
        self.assertFalse(f.is_folder)
        self.assertEqual(f.mime_type, "text/plain")
        self.assertEqual(f.name, "foo")
        self.assertEqual(f.path, "data/foo.txt")

        # Delete it.
        f.delete()
        self.assertFalse(f.exists)

        # Attempt to delete a non-existet file (should do nothing).
        f = File("data/bogus.txt")
        self.assertFalse(f.exists)

        f.delete()
        self.assertFalse(f.exists)



# (C) Copyright 2008-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Version information for this Apptools distribution.

This file is autogenerated by the Apptools setup.py script.
"""

#: The full version of the package, including a development suffix
#: for unreleased versions of the package.
version = "5.0.0"

#: The Git revision from which this release was made.
git_revision = "0e51542048c1aac7ec230c4b773336886bf40792"

#: Flag whether this is a final release
is_released = True



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Naming exceptions. """


class NamingError(Exception):
    """Base class for all naming exceptions."""


class InvalidNameError(NamingError):
    """Invalid name.

    This exception is thrown when the name passed to a naming operation does
    not conform to the syntax of the naming system (or is empty etc).

    """


class NameAlreadyBoundError(NamingError):
    """Name already bound.

    This exception is thrown when an attempt is made to bind a name that is
    already bound in the current context.

    """


class NameNotFoundError(NamingError):
    """Name not found.

    This exception is thrown when a component of a name cannot be resolved
    because it is not bound in the current context.

    """


class NotContextError(NamingError):
    """Not a context.

    This exception is thrown when a naming operation has reached a point where
    a context is required to continue the operation, but the resolved object
    is not a context.

    """


class OperationNotSupportedError(NamingError):
    """The context does support the requested operation."""



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import os
import shutil
import tempfile
import unittest

from traits.api import cached_property, HasTraits, Property, Str, Event

from apptools.naming.api import ObjectSerializer


class FooWithTraits(HasTraits):
    """Dummy HasTraits class for testing ObjectSerizalizer."""

    full_name = Str()

    last_name = Property(depends_on="full_name")

    event = Event()

    @cached_property
    def _get_last_name(self):
        return self.full_name.split(" ")[-1]


class TestObjectSerializer(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)
        self.tmp_file = os.path.join(self.tmpdir, "tmp.pickle")

    def test_save_load_roundtrip(self):
        # Test HasTraits objects can be serialized and deserialized as expected
        obj = FooWithTraits(full_name="John Doe")

        serializer = ObjectSerializer()
        serializer.save(self.tmp_file, obj)

        self.assertTrue(serializer.can_load(self.tmp_file))
        deserialized = serializer.load(self.tmp_file)

        self.assertIsInstance(deserialized, FooWithTraits)
        self.assertEqual(deserialized.full_name, "John Doe")
        self.assertEqual(deserialized.last_name, "Doe")



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from email.mime.multipart import MIMEMultipart
import unittest
from unittest import mock

from apptools.logger.plugin.logger_service import LoggerService


class LoggerServiceTestCase(unittest.TestCase):
    def test_create_email_message(self):
        logger_service = LoggerService()
        with mock.patch.object(
            logger_service, "whole_log_text"
        ) as mocked_log_txt:
            mocked_log_txt.return_value = "Dummy log data"
            msg = logger_service.create_email_message(
                fromaddr="", toaddrs="", ccaddrs="", subject="", priority=""
            )
        self.assertIsInstance(msg, MIMEMultipart)

    def test_create_email_message_with_user_data(self):
        # We used a mocked logger service which doesn't depend on the
        # application trait and the presence of extensions to the extension
        # point `apptools.logger.plugin.mail_files`
        class MockedLoggerService(LoggerService):
            def _get_mail_files(self):
                return [lambda zip_file: None]

        logger_service = MockedLoggerService()
        with mock.patch.object(
            logger_service, "whole_log_text"
        ) as mocked_log_txt:
            mocked_log_txt.return_value = "Dummy log data"
            msg = logger_service.create_email_message(
                fromaddr="",
                toaddrs="",
                ccaddrs="",
                subject="",
                priority="",
                include_userdata=True,
            )
        self.assertIsInstance(msg, MIMEMultipart)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from contextlib import contextmanager

from .file import H5File


@contextmanager
def open_h5file(filename, mode="r+", **kwargs):
    """Context manager for reading an HDF5 file as an H5File object.

    Parameters
    ----------
    filename : str
        HDF5 file name.
    mode : str
        Mode to open the file:

        'r' : Read-only
        'w' : Write; create new file (an existing file would be deleted).
        'a' : Read and write to file; create if not existing
        'r+': Read and write to file; must already exist

    See `H5File` for additional keyword arguments.
    """
    h5 = H5File(filename, mode=mode, **kwargs)
    try:
        yield h5
    finally:
        h5.close()



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# Standard library imports
from io import BytesIO
import logging
import os
import zipfile

# Enthought library imports
from pyface.workbench.api import View as WorkbenchView
from traits.api import (
    Any,
    Callable,
    HasTraits,
    Instance,
    List,
    Property,
    Undefined,
    on_trait_change,
)

root_logger = logging.getLogger()
logger = logging.getLogger(__name__)


class LoggerService(HasTraits):
    """The persistent service exposing the Logger plugin's API."""

    # The Envisage application.
    application = Any()

    # The logging Handler we use.
    handler = Any()

    # Our associated LoggerPreferences.
    preferences = Any()

    # The view we use.
    plugin_view = Instance(WorkbenchView)

    # Contributions from other plugins.
    mail_files = Property(List(Callable))

    def save_preferences(self):
        """Save the preferences."""
        self.preferences.preferences.save()

    def whole_log_text(self):
        """Return all of the logged data as formatted text."""
        lines = [self.handler.format(rec) for rec in self.handler.get()]
        # Ensure that we end with a newline.
        lines.append("")
        text = "\n".join(lines)
        return text

    def create_email_message(
        self,
        fromaddr,
        toaddrs,
        ccaddrs,
        subject,
        priority,
        include_userdata=False,
        stack_trace="",
        comments="",
        include_environment=True,
    ):
        """Format a bug report email from the log files."""
        from email.mime.application import MIMEApplication
        from email.mime.multipart import MIMEMultipart
        from email.mime.text import MIMEText

        message = MIMEMultipart()
        message["Subject"] = "%s [priority=%s]" % (subject, priority)
        message["To"] = ", ".join(toaddrs)
        message["Cc"] = ", ".join(ccaddrs)
        message["From"] = fromaddr
        message.preamble = (
            "You will not see this in a MIME-aware mail " "reader.\n"
        )
        message.epilogue = " "  # To guarantee the message ends with a newline

        # First section is simple ASCII data ...
        m = []
        m.append("Bug Report")
        m.append("==============================")
        m.append("")

        if len(comments) > 0:
            m.append("Comments:")
            m.append("========")
            m.append(comments)
            m.append("")

        if len(stack_trace) > 0:
            m.append("Stack Trace:")
            m.append("===========")
            m.append(stack_trace)
            m.append("")

        msg = MIMEText("\n".join(m))
        message.attach(msg)

        # Include the log file ...
        logtext = self.whole_log_text()
        msg = MIMEText(logtext)
        msg.add_header(
            "Content-Disposition", "attachment", filename="logfile.txt"
        )
        message.attach(msg)

        # Include the environment variables ...
        # FIXME: ask the user, maybe?
        if include_environment:
            # Transmit the user's environment settings as well.  Main purpose
            # is to work out the user name to help with following up on bug
            # reports and in future we should probably send less data.
            entries = []
            for key, value in sorted(os.environ.items()):
                entries.append("%30s : %s\n" % (key, value))

            msg = MIMEText("".join(entries))
            msg.add_header(
                "Content-Disposition", "attachment", filename="environment.txt"
            )
            message.attach(msg)

        if include_userdata and len(self.mail_files) != 0:
            f = BytesIO()
            zf = zipfile.ZipFile(f, "w")
            for mf in self.mail_files:
                mf(zf)
            zf.close()

            msg = MIMEApplication(f.getvalue())
            msg.add_header(
                "Content-Disposition", "attachment", filename="userdata.zip"
            )
            message.attach(msg)

        return message

    def send_bug_report(
        self, smtp_server, fromaddr, toaddrs, ccaddrs, message
    ):
        """Send a bug report email."""
        try:
            import smtplib

            logger.debug("Connecting to: %s" % smtp_server)
            server = smtplib.SMTP(host=smtp_server)
            logger.debug("Connected: %s" % server)
            # server.set_debuglevel(1)
            server.sendmail(fromaddr, toaddrs + ccaddrs, message.as_string())
            server.quit()
        except Exception:
            logger.exception("Problem sending error report")

    #### Traits stuff #########################################################

    def _get_mail_files(self):
        return self.application.get_extensions(
            "apptools.logger.plugin.mail_files"
        )

    @on_trait_change("preferences.level_")
    def _level_changed(self, new):
        if (
            new is not None
            and new is not Undefined
            and self.handler is not None
        ):
            root_logger.setLevel(self.preferences.level_)
            self.handler.setLevel(self.preferences.level_)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This was previously a test for the now deleted apptools.sweet_pickle
sub package.  It is included here to showcase how apptools.persistance can be
used to replace sweet_pickle functionality.
"""

import io
import random
import re
import pickle
import unittest

from apptools.persistence.versioned_unpickler import VersionedUnpickler

########################################


# Usecase1: generic case
class A(object):
    def __init__(self, b=None):
        self.x = 0
        self.set_b(b)

    def set_b(self, b):
        self.b_ref = b
        if b and hasattr(b, "y"):
            self.x = b.y

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.set_b(self.b_ref)

    def __initialize__(self):
        while self.b_ref is None and self.b_ref.y != 0:
            yield True
        self.set_b(self.b_ref)


class B(object):
    def __init__(self, a=None):
        self.y = 0
        self.set_a(a)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["y"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.set_a(self.a_ref)

    def set_a(self, a):
        self.a_ref = a
        if a and hasattr(a, "x"):
            self.y = a.x

    def __initialize__(self):
        while self.a_ref is None and self.a_ref.x != 0:
            yield True
        self.set_a(self.a_ref)


class GenericTestCase(unittest.TestCase):
    def test_generic(self):
        a = A()
        b = B()
        a.x = random.randint(1, 100)
        b.set_a(a)
        a.set_b(b)
        value = a.x

        # This will fail, even though we have a __setstate__ method.
        s = pickle.dumps(a)
        new_a = pickle.loads(s)
        # Accessing new_a.x is okay
        new_a.x
        # Accessing y directly would fail
        with self.assertRaisesRegex(
                AttributeError, "'B' object has no attribute 'y'"):
            new_a.b_ref.y

        # This will work!
        s = pickle.dumps(a)
        new_a = VersionedUnpickler(io.BytesIO(s)).load()
        assert new_a.x == new_a.b_ref.y == value


########################################
# Usecase2: Toy Application


class StringFinder(object):
    def __init__(self, source, pattern):
        self.pattern = pattern
        self.source = source
        self.data = []

    def __getstate__(self):
        s = self.__dict__.copy()
        del s["data"]
        return s

    def __initialize__(self):
        while not self.source.initialized:
            yield True
        self.find()

    def find(self):
        pattern = self.pattern
        string = self.source.data
        self.data = [
            (x.start(), x.end()) for x in re.finditer(pattern, string)
        ]


class XMLFileReader(object):
    def __init__(self, file_name):
        self.data = ""
        self.initialized = False
        self.file_name = file_name
        self.read()

    def __getstate__(self):
        s = self.__dict__.copy()
        del s["data"]
        del s["initialized"]
        return s

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.read()

    def read(self):
        # Make up random data from the filename
        data = [10 * x for x in self.file_name]
        random.shuffle(data)
        self.data = " ".join(data)
        self.initialized = True


class Application(object):
    def __init__(self):
        self.reader = XMLFileReader("some_test_file.xml")
        self.finder = StringFinder(self.reader, "e")

    def get(self):
        return (self.finder.data, self.reader.data)


class ToyAppTestCase(unittest.TestCase):
    def test_toy_app(self):
        a = Application()
        a.finder.find()
        a.get()
        s = pickle.dumps(a)
        b = pickle.loads(s)

        with self.assertRaisesRegex(
                AttributeError,
                "'StringFinder' object has no attribute 'data'"):
            b.get()

        # Works fine.
        c = VersionedUnpickler(io.BytesIO(s)).load()
        c.get()



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from traits.api import Interface, List, Str


class ISelection(Interface):
    """ Collection of selected items. """

    #: ID of the selection provider that created this selection object.
    provider_id = Str

    def is_empty(self):
        """ Is the selection empty? """


class IListSelection(ISelection):
    """ Selection for ordered sequences of items. """

    #: Selected objects.
    items = List

    #: Indices of the selected objects in the selection provider.
    indices = List



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from contextlib import contextmanager
import tempfile
import os

from ..utils import open_h5file


SEPARATOR = "-" * 60


@contextmanager
def temp_file(suffix="", prefix="tmp", dir=None):
    fd, filename = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        yield filename
    finally:
        os.close(fd)
        os.unlink(filename)


@contextmanager
def temp_h5_file(**kwargs):
    with temp_file(suffix="h5") as fn:
        with open_h5file(fn, mode="a", **kwargs) as h5:
            yield h5



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests for scoped preferences. """


# Standard library imports.
import os
import tempfile
from os.path import join

# Major package imports.
from importlib_resources import files

# Enthought library imports.
from apptools.preferences.api import Preferences, ScopedPreferences

# Local imports.
from .test_preferences import PreferencesTestCase


# This module's package.
PKG = "apptools.preferences.tests"


class ScopedPreferencesTestCase(PreferencesTestCase):
    """ Tests for the scoped preferences. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.preferences = ScopedPreferences()

        # The filename of the example preferences file.
        self.example = os.fspath(files(PKG) / "example.ini")

        # A temporary directory that can safely be written to.
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        # Remove the temporary directory.
        os.rmdir(self.tmpdir)

    ###########################################################################
    # Tests overridden from 'PreferencesTestCase'.
    ###########################################################################

    def test_node(self):
        """ node """

        p = self.preferences

        # Try an empty path.
        self.assertEqual(p, p.node())

        # Try a simple path.
        node = p.node("acme")
        self.assertIsNotNone(node)
        self.assertEqual("acme", node.name)
        self.assertEqual("acme", node.path)
        self.assertEqual(p.node("application/"), node.parent)

        # Make sure we get the same node each time we ask for it!
        self.assertEqual(node, p.node("acme"))

        # Try a nested path.
        node = p.node("acme.ui")
        self.assertIsNotNone(node)
        self.assertEqual("ui", node.name)
        self.assertEqual("acme.ui", node.path)
        self.assertEqual(p.node("application/acme"), node.parent)

        # And just to be sure, a really nested path.
        node = p.node("acme.ui.splash_screen")
        self.assertIsNotNone(node)
        self.assertEqual("splash_screen", node.name)
        self.assertEqual("acme.ui.splash_screen", node.path)
        self.assertEqual(p.node("application/acme.ui"), node.parent)

    def test_save(self):
        """ save """

        p = self.preferences

        # Get the application scope.
        application = p.node("application/")

        tmp = join(self.tmpdir, "test.ini")
        application.filename = tmp

        # Set a value.
        p.set("acme.ui.bgcolor", "red")

        # Save all scopes.
        p.save()

        # Make sure a file was written.
        self.assertTrue(os.path.exists(tmp))

        # Load the 'ini' file into a new preferences node and make sure the
        # preference is in there.
        p = Preferences()
        p.load(tmp)

        self.assertEqual("red", p.get("acme.ui.bgcolor"))

        # Cleanup.
        os.remove(tmp)

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_ability_to_specify_primary_scope(self):

        preferences = ScopedPreferences(
            scopes=[
                Preferences(name="a"),
                Preferences(name="b"),
                Preferences(name="c"),
            ],
            primary_scope_name="b",
        )

        # This should set the prefrrence in the primary scope.
        preferences.set("acme.foo", "bar")

        # Look it up specifically in the primary scope.
        self.assertEqual("bar", preferences.get("b/acme.foo"))

    def test_builtin_scopes(self):
        """ builtin scopes """

        p = self.preferences

        # Make sure the default built-in scopes get created.
        self.assertTrue(p.node_exists("application/"))
        self.assertTrue(p.node_exists("default/"))

    def test_get_and_set_in_specific_scope(self):
        """ get and set in specific scope """

        p = self.preferences

        # Set a preference and make sure we can get it again!
        p.set("default/acme.ui.bgcolor", "red")
        self.assertEqual("red", p.get("default/acme.ui.bgcolor"))

    def test_clear_in_specific_scope(self):
        """ clear in specific scope """

        p = self.preferences

        # Set a value in both the application and default scopes.
        p.set("application/acme.ui.bgcolor", "red")
        p.set("default/acme.ui.bgcolor", "yellow")

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual("red", p.get("acme.ui.bgcolor"))

        # Now clear out the application scope.
        p.clear("application/acme.ui")
        self.assertEqual(0, len(p.keys("application/acme.ui")))

        # We should now get the value from the default scope.
        self.assertEqual("yellow", p.get("acme.ui.bgcolor"))

    def test_remove_in_specific_scope(self):
        """ remove in specific scope """

        p = self.preferences

        # Set a value in both the application and default scopes.
        p.set("application/acme.ui.bgcolor", "red")
        p.set("default/acme.ui.bgcolor", "yellow")

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual("red", p.get("acme.ui.bgcolor"))

        # Now remove it from the application scope.
        p.remove("application/acme.ui.bgcolor")

        # We should now get the value from the default scope.
        self.assertEqual("yellow", p.get("acme.ui.bgcolor"))

    def test_keys_in_specific_scope(self):
        """ keys in specific scope """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.keys("default/"))

        # Set some preferences in the node.
        p.set("default/a", "1")
        p.set("default/b", "2")
        p.set("default/c", "3")

        keys = p.keys("default/")
        keys.sort()

        self.assertEqual(["a", "b", "c"], keys)

        # Set some preferences in a child node.
        p.set("default/acme.a", "1")
        p.set("default/acme.b", "2")
        p.set("default/acme.c", "3")

        keys = p.keys("default/acme")
        keys.sort()

        self.assertEqual(["a", "b", "c"], keys)

        # And, just to be sure, in a child of the child node ;^)
        p.set("default/acme.ui.a", "1")
        p.set("default/acme.ui.b", "2")
        p.set("default/acme.ui.c", "3")

        keys = p.keys("default/acme.ui")
        keys.sort()

        self.assertEqual(["a", "b", "c"], keys)

    def test_node_in_specific_scope(self):
        """ node in specific scope """

        p = self.preferences

        # Try an empty path.
        self.assertEqual(p, p.node())

        # Try a simple path.
        node = p.node("default/acme")
        self.assertIsNotNone(node)
        self.assertEqual("acme", node.name)
        self.assertEqual("acme", node.path)
        self.assertEqual(p.node("default/"), node.parent)

        # Make sure we get the same node each time we ask for it!
        self.assertEqual(node, p.node("default/acme"))

        # Try a nested path.
        node = p.node("default/acme.ui")
        self.assertIsNotNone(node)
        self.assertEqual("ui", node.name)
        self.assertEqual("acme.ui", node.path)
        self.assertEqual(p.node("default/acme"), node.parent)

        # And just to be sure, a really nested path.
        node = p.node("default/acme.ui.splash_screen")
        self.assertIsNotNone(node)
        self.assertEqual("splash_screen", node.name)
        self.assertEqual("acme.ui.splash_screen", node.path)
        self.assertEqual(p.node("default/acme.ui"), node.parent)

    def test_node_exists_in_specific_scope(self):
        """ node exists """

        p = self.preferences

        self.assertTrue(p.node_exists())
        self.assertFalse(p.node_exists("default/acme"))

        p.node("default/acme")
        self.assertTrue(p.node_exists("default/acme"))

    def test_node_names_in_specific_scope(self):
        """ node names in specific scope """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.node_names("default/"))

        # Create some nodes.
        p.node("default/a")
        p.node("default/b")
        p.node("default/c")

        names = p.node_names("default/")
        names.sort()

        self.assertEqual(["a", "b", "c"], names)

        # Creatd some nodes in a child node.
        p.node("default/acme.a")
        p.node("default/acme.b")
        p.node("default/acme.c")

        names = p.node_names("default/acme")
        names.sort()

        self.assertEqual(["a", "b", "c"], names)

        # And, just to be sure, in a child of the child node ;^)
        p.node("default/acme.ui.a")
        p.node("default/acme.ui.b")
        p.node("default/acme.ui.c")

        names = p.node_names("default/acme.ui")
        names.sort()

        self.assertEqual(["a", "b", "c"], names)

    def test_default_lookup_order(self):
        """ default lookup order """

        p = self.preferences

        # Set a value in both the application and default scopes.
        p.set("application/acme.ui.bgcolor", "red")
        p.set("default/acme.ui.bgcolor", "yellow")

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual("red", p.get("acme.ui.bgcolor"))

        # But we can still get at each scope individually.
        self.assertEqual("red", p.get("application/acme.ui.bgcolor"))
        self.assertEqual("yellow", p.get("default/acme.ui.bgcolor"))

    def test_lookup_order(self):
        """ lookup order """

        p = self.preferences
        p.lookup_order = ["default", "application"]

        # Set a value in both the application and default scopes.
        p.set("application/acme.ui.bgcolor", "red")
        p.set("default/acme.ui.bgcolor", "yellow")

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual("red", p.get("acme.ui.bgcolor"))

        # But we can still get at each scope individually.
        self.assertEqual("red", p.get("application/acme.ui.bgcolor"))
        self.assertEqual("yellow", p.get("default/acme.ui.bgcolor"))

    def test_add_listener_in_specific_scope(self):
        """ add listener in specific scope. """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key = key
            listener.old = old
            listener.new = new

        # Add a listener.
        p.add_preferences_listener(listener, "default/acme.ui")

        # Set a value and make sure the listener was called.
        p.set("default/acme.ui.bgcolor", "blue")
        self.assertEqual(p.node("default/acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertIsNone(listener.old)
        self.assertEqual("blue", listener.new)

        # Set it to another value to make sure we get the 'old' value
        # correctly.
        p.set("default/acme.ui.bgcolor", "red")
        self.assertEqual(p.node("default/acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertEqual("blue", listener.old)
        self.assertEqual("red", listener.new)

    def test_remove_listener_in_specific_scope(self):
        """ remove listener in specific scope. """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key = key
            listener.old = old
            listener.new = new

        # Add a listener.
        p.add_preferences_listener(listener, "default/acme.ui")

        # Set a value and make sure the listener was called.
        p.set("default/acme.ui.bgcolor", "blue")
        self.assertEqual(p.node("default/acme.ui"), listener.node)
        self.assertEqual("bgcolor", listener.key)
        self.assertIsNone(listener.old)
        self.assertEqual("blue", listener.new)

        # Remove the listener.
        p.remove_preferences_listener(listener, "default/acme.ui")

        # Set a value and make sure the listener was *not* called.
        listener.node = None

        p.set("default/acme.ui.bgcolor", "blue")
        self.assertIsNone(listener.node)

    def test_non_existent_scope(self):
        """ non existent scope """

        p = self.preferences

        self.assertRaises(ValueError, p.get, "bogus/acme.ui.bgcolor")



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


def __replacement_setstate__(self, state):
    """"""
    state = self.__updater__(state)
    self.__dict__.update(state)


class Updater:

    """An abstract class to provide functionality common to the updaters."""

    def get_latest(self, module, name):
        """The refactorings dictionary contains mappings between old and new
        module names.  Since we only bump the version number one increment
        there is only one possible answer.
        """
        if hasattr(self, "refactorings"):
            module = self.strip(module)
            name = self.strip(name)
            # returns the new module and name if it exists otherwise defaults
            # to using the original module and name
            module, name = self.refactorings.get(
                (module, name), (module, name)
            )

        return module, name

    def strip(self, string):
        # Who would have thought that pickle would pass us
        # names with \013 on the end? Is this after the files have
        # manually edited?
        if ord(string[-1:]) == 13:
            return string[:-1]

        return string



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""
A Recorder subclass that presents a simple user interface.
"""

from traits.api import Code, Button, Int, on_trait_change, Any
from traitsui.api import View, Item, Group, HGroup, CodeEditor, spring, Handler

from .recorder import Recorder


###############################################################################
# `CloseHandler` class.
###############################################################################
class CloseHandler(Handler):
    """This class cleans up after the UI for the recorder is closed."""

    def close(self, info, is_ok):
        """This method is invoked when the user closes the UI."""
        recorder = info.object
        recorder.on_ui_close()
        return True


###############################################################################
# `RecorderWithUI` class.
###############################################################################
class RecorderWithUI(Recorder):
    """
    This class represents a Recorder but with a simple user interface.
    """

    # The code to display
    code = Code(editor=CodeEditor(line="current_line"))

    # Button to save script to file.
    save_script = Button("Save Script")

    # The current line to show, used by the editor.
    current_line = Int

    # The root object which is being recorded.
    root = Any

    ########################################
    # Traits View.
    view = View(
        Group(
            HGroup(
                Item("recording", show_label=True),
                spring,
                Item("save_script", show_label=False),
            ),
            Group(Item("code", show_label=False)),
        ),
        width=600,
        height=360,
        id="apptools.scripting.recorder_with_ui",
        buttons=["Cancel"],
        resizable=True,
        handler=CloseHandler(),
    )

    ######################################################################
    # RecorderWithUI interface.
    ######################################################################
    def on_ui_close(self):
        """Called from the CloseHandler when the UI is closed. This
        method basically stops the recording.
        """
        from .util import stop_recording
        from .package_globals import get_recorder

        if get_recorder() is self:
            stop_recording(self.root, save=False)
        else:
            self.recording = False
            self.unregister(self.root)

    ######################################################################
    # Non-public interface.
    ######################################################################
    @on_trait_change("lines[]")
    def _update_code(self):
        self.code = self.get_code()
        self.current_line = len(self.lines) + 1

    def _save_script_fired(self):
        self.ui_save()



# (C) Copyright 2006-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Utilities for handling optional dependencies so that tests for
optional features can be skipped gracefully.
"""
import importlib
import unittest


def optional_import(name):
    """
    Optionally import a module, returning None if that module is unavailable.

    Parameters
    ----------
    name : Str
        The name of the module being imported.

    Returns
    -------
    None or module
        None if the module is not available, and the module otherwise.

    """
    try:
        module = importlib.import_module(name)
    except ImportError:
        return None
    else:
        return module


numpy = optional_import("numpy")
requires_numpy = unittest.skipIf(numpy is None, "NumPy not available")

pandas = optional_import("pandas")
requires_pandas = unittest.skipIf(pandas is None, "Pandas not available")

tables = optional_import("tables")
requires_tables = unittest.skipIf(tables is None, "PyTables not available")



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Convenience functions for creating logging handlers etc. """


# Standard library imports.
import logging
from logging.handlers import RotatingFileHandler

# Local imports.
from .log_queue_handler import LogQueueHandler


# The default logging level.
LEVEL = logging.DEBUG

# The default formatter.
FORMATTER = logging.Formatter("%(levelname)s|%(asctime)s|%(message)s")


class LogFileHandler(RotatingFileHandler):
    """The default log file handler."""

    def __init__(
        self, path, maxBytes=1000000, backupCount=3, level=None, formatter=None
    ):
        RotatingFileHandler.__init__(
            self, path, maxBytes=maxBytes, backupCount=3
        )

        if level is None:
            level = LEVEL
        if formatter is None:
            formatter = FORMATTER
        # Set our default formatter and log level.
        self.setFormatter(formatter)
        self.setLevel(level)


def add_log_queue_handler(logger, level=None, formatter=None):
    """Adds a queueing log handler to a logger."""
    if level is None:
        level = LEVEL
    if formatter is None:
        formatter = FORMATTER

    # Add the handler to the root logger.
    log_queue_handler = LogQueueHandler()
    log_queue_handler.setLevel(level)
    log_queue_handler.setFormatter(formatter)
    logger.addHandler(log_queue_handler)
    return log_queue_handler



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests for the preferences page. """

import unittest

from traits.api import (
    Enum, List, Str,
    pop_exception_handler,
    push_exception_handler,
)
from traitsui.api import Group, Item, View

from apptools.preferences.api import Preferences
from apptools.preferences.ui.api import PreferencesPage


class TestPreferencesPage(unittest.TestCase):
    """ Non-GUI Tests for PreferencesPage."""

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_preferences_page_apply(self):
        """ Test applying the preferences """

        # this sets up imitate Mayavi usage.

        class MyPreferencesPage(PreferencesPage):

            # the following set default values for class traits
            category = "Application"

            help_id = ""

            name = "Note"

            preferences_path = "my_ref.pref"

            # custom preferences

            backend = Enum("auto", "simple", "test")

            traits_view = View(Group(Item("backend")))

        preferences = Preferences()
        pref_page = MyPreferencesPage(
            preferences=preferences,
            category="Another Application",
            help_id="this_wont_be_saved",
            name="Different Note",
            # custom preferences
            backend="simple",
        )
        pref_page.apply()

        self.assertEqual(preferences.get("my_ref.pref.backend"), "simple")
        self.assertEqual(preferences.keys("my_ref.pref"), ["backend"])

        # this is not saved by virtue of it being static and never assigned to
        self.assertIsNone(preferences.get("my_ref.pref.traits_view"))

        # These are skipped because this trait is defined on the
        # PreferencesPage.
        self.assertIsNone(preferences.get("my_ref.pref.help_id"))
        self.assertIsNone(preferences.get("my_ref.pref.category"))
        self.assertIsNone(preferences.get("my_ref.pref.name"))

    def test_preferences_page_apply_skip_items_traits(self):
        """ Test _items traits from List mutation are skipped. """
        # Regression test for enthought/apptools#129

        class MyPreferencesPage(PreferencesPage):
            preferences_path = "my_ref.pref"
            names = List(Str())

        preferences = Preferences()
        pref_page = MyPreferencesPage(
            preferences=preferences,
            names=["1"],
        )
        pref_page.names.append("2")
        pref_page.apply()

        self.assertEqual(preferences.get("my_ref.pref.names"), str(["1", "2"]))
        self.assertEqual(preferences.keys("my_ref.pref"), ["names"])

    def test_sync_anytrait_items_overload(self):
        """ Test sychronizing trait with name *_items not to be mistaken
        as the event trait for mutating list/dict/set
        """

        class MyPreferencesPage(PreferencesPage):
            preferences_path = Str('my_section')

            names_items = Str()

        preferences = Preferences()
        pref_page = MyPreferencesPage(preferences=preferences)
        pref_page.names_items = "Hello"
        pref_page.apply()

        self.assertEqual(
            sorted(preferences.keys("my_section")),
            ["names_items"]
        )
        self.assertEqual(
            preferences.get("my_section.names_items"),
            "Hello",
        )



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from ..type_registry import TypeRegistry

from .dummies import A, B, C, D, Mixed, Abstract, Concrete, ConcreteSubclass


class TestTypeRegistry(unittest.TestCase):
    def setUp(self):
        self.registry = TypeRegistry()

    def test_deferred_push(self):
        self.registry.push("dummies:A", "A")
        self.registry.push("dummies:C", "C")
        self.assertEqual(self.registry.lookup_by_type(A), "A")
        self.assertEqual(self.registry.lookup_by_type(B), "A")
        self.assertEqual(self.registry.lookup_by_type(C), "C")
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_greedy_push(self):
        self.registry.push(A, "A")
        self.registry.push(C, "C")
        self.assertEqual(self.registry.lookup_by_type(A), "A")
        self.assertEqual(self.registry.lookup_by_type(B), "A")
        self.assertEqual(self.registry.lookup_by_type(C), "C")
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_pop_by_name_from_name(self):
        self.registry.push("dummies:A", "A")
        self.registry.pop("dummies:A")
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_pop_by_name_from_type(self):
        self.registry.push(A, "A")
        self.registry.pop("dummies:A")
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_pop_by_type_from_name(self):
        self.registry.push("dummies:A", "A")
        self.registry.pop(A)
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_pop_by_type_from_type(self):
        self.registry.push(A, "A")
        self.registry.pop(A)
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_mro(self):
        self.registry.push("dummies:A", "A")
        self.registry.push("dummies:D", "D")
        self.assertEqual(self.registry.lookup_by_type(Mixed), "A")

    def test_lookup_instance(self):
        self.registry.push(A, "A")
        self.registry.push(C, "C")
        self.assertEqual(self.registry.lookup(A()), "A")
        self.assertEqual(self.registry.lookup(B()), "A")
        self.assertEqual(self.registry.lookup(C()), "C")
        self.assertRaises(KeyError, self.registry.lookup, D())

    def test_lookup_all(self):
        self.registry.push(A, "A")
        self.registry.push(C, "C")
        self.assertEqual(self.registry.lookup_all(A()), ["A"])
        self.assertEqual(self.registry.lookup_all(B()), ["A"])
        self.registry.push(A, "A2")
        self.assertEqual(self.registry.lookup_all(A()), ["A", "A2"])
        self.assertEqual(self.registry.lookup_all(B()), ["A", "A2"])

    def test_abc(self):
        self.registry.push_abc(Abstract, "Abstract")
        self.assertEqual(self.registry.lookup_by_type(Concrete), "Abstract")
        self.assertEqual(
            self.registry.lookup_by_type(ConcreteSubclass), "Abstract"
        )

    def test_stack_type(self):
        self.registry.push(A, "A1")
        self.registry.push(A, "A2")
        self.assertEqual(self.registry.lookup_by_type(A), "A2")
        self.registry.pop(A)
        self.assertEqual(self.registry.lookup_by_type(A), "A1")
        self.registry.pop(A)
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)
        self.assertRaises(KeyError, self.registry.pop, A)
        self.assertRaises(KeyError, self.registry.pop, "dummies:A")
        self.assertEqual(self.registry.type_map, {})
        self.assertEqual(self.registry.name_map, {})
        self.assertEqual(self.registry.abc_map, {})

    def test_stack_mixed_type_and_name(self):
        self.registry.push(A, "A1")
        self.registry.push("dummies:A", "A2")
        self.assertEqual(self.registry.lookup_by_type(A), "A2")
        self.registry.pop(A)
        self.assertEqual(self.registry.lookup_by_type(A), "A1")
        self.registry.pop("dummies:A")
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)
        self.assertRaises(KeyError, self.registry.pop, A)
        self.assertRaises(KeyError, self.registry.pop, "dummies:A")
        self.assertEqual(self.registry.type_map, {})
        self.assertEqual(self.registry.name_map, {})
        self.assertEqual(self.registry.abc_map, {})

    def test_stack_abc(self):
        self.registry.push_abc(Abstract, "Abstract1")
        self.registry.push_abc(Abstract, "Abstract2")
        self.assertEqual(self.registry.lookup_by_type(Concrete), "Abstract2")
        self.registry.pop(Abstract)
        self.assertEqual(self.registry.lookup_by_type(Concrete), "Abstract1")
        self.registry.pop(Abstract)
        self.assertRaises(KeyError, self.registry.lookup_by_type, Concrete)
        self.assertRaises(KeyError, self.registry.pop, Abstract)
        self.assertEqual(self.registry.type_map, {})
        self.assertEqual(self.registry.name_map, {})
        self.assertEqual(self.registry.abc_map, {})



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Object factory for Python File System contexts. """


# Enthought library imports.
from apptools.io.api import File

# Local imports.
from .object_factory import ObjectFactory
from .reference import Reference


class PyFSObjectFactory(ObjectFactory):
    """ Object factory for Python File System contexts. """

    ###########################################################################
    # 'ObjectFactory' interface.
    ###########################################################################

    def get_object_instance(self, state, name, context):
        """ Creates an object using the specified state information. """

        obj = None

        if isinstance(state, Reference):
            if state.class_name == "File" and len(state.addresses) > 0:
                obj = File(state.addresses[0].content)

        return obj



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import os
from contextlib import closing
import unittest

from apptools._testing.optional_dependencies import (
    numpy as np,
    tables,
    requires_numpy,
    requires_tables,
)

if np is not None and tables is not None:
    from ..file import H5File
    from ..dict_node import H5DictNode
    from ..table_node import H5TableNode
    from .utils import open_h5file, temp_h5_file


H5_TEST_FILE = "_temp_test_filt.h5"


@requires_numpy
@requires_tables
class FileTestCase(unittest.TestCase):
    def tearDown(self):
        try:
            os.remove(H5_TEST_FILE)
        except OSError:
            pass

    def test_reopen(self):
        h5 = H5File(H5_TEST_FILE, mode="w")
        assert h5.is_open
        h5.close()
        assert not h5.is_open
        h5.open()
        assert h5.is_open
        h5.close()

    def test_open_from_pytables_object(self):
        with closing(tables.File(H5_TEST_FILE, "w")) as pyt_file:
            pyt_file.create_group("/", "my_group")
            with open_h5file(pyt_file) as h5:
                assert "/my_group" in h5

    def test_open_from_closed_pytables_object(self):
        with closing(tables.File(H5_TEST_FILE, "w")) as pyt_file:
            pyt_file.create_group("/", "my_group")
            pyt_file.close()
            with open_h5file(pyt_file) as h5:
                assert "/my_group" in h5

    def test_create_array_with_H5File(self):
        array = np.arange(3)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.create_array("/array", array)
            # Test returned array
            np.testing.assert_allclose(h5array, array)
            # Test stored array
            np.testing.assert_allclose(h5["/array"], array)

    def test_create_array_with_H5Group(self):
        array = np.arange(3)
        node_path = "/tardigrade/array"
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            group = h5.create_group("/tardigrade")
            h5array = group.create_array("array", array)
            # Test returned array
            np.testing.assert_allclose(h5array, array)
            # Test stored array
            np.testing.assert_allclose(h5[node_path], array)

    def test_getitem_failure(self):
        array = np.arange(3)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_array("/array", array)
            with self.assertRaises(NameError):
                h5["/not_there"]

    def test_iteritems(self):
        node_paths = ["/foo", "/bar", "/baz"]
        array = np.arange(3)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            for path in node_paths:
                h5.create_array(path, array)

            # We expect to see the root node when calling iteritems...
            node_paths.append("/")
            iter_paths = []

            # 2to3 converts the iteritems blindly to items which is incorrect,
            # so we resort to this ugliness.
            items = getattr(h5, "iteritems")()
            for path, node in items:
                iter_paths.append(path)

        assert set(node_paths) == set(iter_paths)

    def test_create_plain_array_with_H5File(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.create_array("/array", np.arange(3), chunked=False)
            assert isinstance(h5array, tables.Array)
            assert not isinstance(h5array, tables.CArray)

    def test_create_plain_array_with_H5Group(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.root.create_array(
                "/array", np.arange(3), chunked=False
            )
            assert isinstance(h5array, tables.Array)
            assert not isinstance(h5array, tables.CArray)

    def test_create_chunked_array_with_H5File(self):
        array = np.arange(3, dtype=np.uint8)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.create_array("/array", array, chunked=True)
            np.testing.assert_allclose(h5array, array)
            assert isinstance(h5array, tables.CArray)

    def test_create_chunked_array_with_H5Group(self):
        array = np.arange(3, dtype=np.uint8)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.root.create_array("/array", array, chunked=True)
            np.testing.assert_allclose(h5array, array)
            assert isinstance(h5array, tables.CArray)

    def test_create_extendable_array_with_H5File(self):
        array = np.arange(3, dtype=np.uint8)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.create_array("/array", array, extendable=True)
            np.testing.assert_allclose(h5array, array)
            assert isinstance(h5array, tables.EArray)

    def test_create_extendable_array_with_H5Group(self):
        array = np.arange(3, dtype=np.uint8)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5array = h5.root.create_array("/array", array, extendable=True)
            np.testing.assert_allclose(h5array, array)
            assert isinstance(h5array, tables.EArray)

    def test_str_and_repr(self):
        array = np.arange(3)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_array("/array", array)

            assert repr(h5) == repr(h5._h5)
            assert str(h5) == str(h5._h5)

    def test_shape_and_dtype(self):
        array = np.ones((3, 4), dtype=np.uint8)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            for node, chunked in (("/array", False), ("/carray", True)):
                h5array = h5.create_array(
                    node, array.shape, dtype=array.dtype, chunked=chunked
                )
                assert h5array.dtype == array.dtype
                assert h5array.shape == array.shape

    def test_shape_only_raises(self):
        shape = (3, 4)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            with self.assertRaises(ValueError):
                h5.create_array("/array", shape)

    def test_create_duplicate_array_raises(self):
        array = np.arange(3)
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=False) as h5:
            h5.create_array("/array", array)
            with self.assertRaises(ValueError):
                h5.create_array("/array", array)

    def test_delete_existing_array_with_H5File(self):
        old_array = np.arange(3)
        new_array = np.ones(5)
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            h5.create_array("/array", old_array)
            # New array with the same node name should delete old array
            h5.create_array("/array", new_array)
            np.testing.assert_allclose(h5["/array"], new_array)

    def test_delete_existing_array_with_H5Group(self):
        old_array = np.arange(3)
        new_array = np.ones(5)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_array("/array", old_array)
            # New array with the same node name should delete old array
            h5.root.create_array("/array", new_array, delete_existing=True)
            np.testing.assert_allclose(h5["/array"], new_array)

    def test_delete_existing_dict_with_H5File(self):
        old_dict = {"a": "Goose"}
        new_dict = {"b": "Quail"}
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            h5.create_dict("/dict", old_dict)
            # New dict with the same node name should delete old dict
            h5.create_dict("/dict", new_dict)
            assert h5["/dict"].data == new_dict

    def test_delete_existing_dict_with_H5Group(self):
        old_dict = {"a": "Goose"}
        new_dict = {"b": "Quail"}
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_dict("/dict", old_dict)
            # New dict with the same node name should delete old dict
            h5.root.create_dict("/dict", new_dict, delete_existing=True)
            assert h5["/dict"].data == new_dict

    def test_delete_existing_table_with_H5File(self):
        old_description = [("Honk", "int"), ("Wink", "float")]
        new_description = [("Toot", "float"), ("Pop", "int")]
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            h5.create_table("/table", old_description)
            # New table with the same node name should delete old table
            h5.create_table("/table", new_description)
            tab = h5["/table"]
            tab.append({"Pop": (1,), "Toot": (np.pi,)})
            assert tab.ix[0][0] == np.pi

    def test_delete_existing_table_with_H5Group(self):
        old_description = [("Honk", "int"), ("Wink", "float")]
        new_description = [("Toot", "float"), ("Pop", "int")]
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_table("/table", old_description)
            # New table with the same node name should delete old table
            h5.root.create_table(
                "/table", new_description, delete_existing=True
            )
            tab = h5["/table"]
            tab.append({"Pop": (1,), "Toot": (np.pi,)})
            assert tab.ix[0][0] == np.pi

    def test_delete_existing_group_with_H5File(self):
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            h5.create_group("/group")
            grp = h5["/group"]
            grp.attrs["test"] = 4

            assert grp.attrs["test"] == 4

            h5.create_group("/group")
            grp = h5["/group"]
            grp.attrs["test"] = 6

            assert grp.attrs["test"] == 6

    def test_delete_existing_group_with_H5Group(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_group("/group")
            grp = h5["/group"]
            grp.attrs["test"] = 4

            assert grp.attrs["test"] == 4

            h5.root.create_group("/group", delete_existing=True)
            grp = h5["/group"]
            grp.attrs["test"] = 6

            assert grp.attrs["test"] == 6

    def test_remove_group_with_H5File(self):
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            h5.create_group("/group")
            assert "/group" in h5
            h5.remove_group("/group")
            assert "/group" not in h5

    def test_remove_group_with_H5Group(self):
        node_path = "/waterbear/group"
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            group = h5.create_group("/waterbear")
            group.create_group("group")
            assert node_path in h5
            group.remove_group("group")
            assert node_path not in h5

    def test_remove_group_with_remove_node(self):
        node_path = "/group"
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_group(node_path)
            with self.assertRaises(ValueError):
                # Groups should be removed w/ `remove_group`
                h5.remove_node(node_path)

    def test_remove_node_with_H5File(self):
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            h5.create_array("/array", np.arange(3))
            assert "/array" in h5
            h5.remove_node("/array")
            assert "/array" not in h5

    def test_remove_node_with_H5Group(self):
        node_path = "/waterbear/array"
        with open_h5file(H5_TEST_FILE, mode="w", delete_existing=True) as h5:
            group = h5.create_group("/waterbear")
            h5.create_array(node_path, np.arange(3))
            assert node_path in h5
            group.remove_node("array")
            assert node_path not in h5

    def test_read_mode_raises_on_nonexistent_file(self):
        cm = open_h5file("_nonexistent_.h5", mode="r")
        with self.assertRaises(IOError):
            cm.__enter__()

        cm = open_h5file("_nonexistent_.h5", mode="r+")
        with self.assertRaises(IOError):
            cm.__enter__()

    def test_cleanup(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5_pytables = h5._h5  # This reference gets deleted on close
        assert not h5_pytables.isopen

    def test_create_group_with_H5File(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_group("/group")
            assert "/group" in h5

    def test_create_group_with_H5Group(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            group = h5["/"]
            group.create_group("group")
            assert "/group" in h5

    def test_split_path(self):
        path, node = H5File.split_path("/")
        assert path == "/"
        assert node == ""

        path, node = H5File.split_path("/node")
        assert path == "/"
        assert node == "node"

        path, node = H5File.split_path("/group/node")
        assert path == "/group"
        assert node == "node"

    def test_join_path(self):
        path = H5File.join_path("/", "a", "b", "c")
        assert path == "/a/b/c"

        path = H5File.join_path("a", "b/c")
        assert path == "/a/b/c"

        path = H5File.join_path("a", "/b", "/c")
        assert path == "/a/b/c"

    def test_auto_groups(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.auto_groups = True
            h5.create_array("/group/array", np.arange(3))
            assert "/group/array" in h5

    def test_auto_groups_deep(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.auto_groups = True
            h5.create_array("/group1/group2/array", np.arange(3))
            assert "/group1/group2/array" in h5

    def test_groups(self):
        array = np.arange(3)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5.create_array("/a/b/array", array)
            group_a = h5["/a"]
            # Check that __contains__ works for groups
            assert "b" in group_a
            assert "b/array" in group_a
            group_b = group_a["b"]
            # Check that __contains__ works for arrays
            assert "array" in group_b
            np.testing.assert_allclose(group_b["array"], array)
            np.testing.assert_allclose(group_a["b/array"], array)

    def test_group_attributes(self):
        value_1 = "foo"
        value_2 = "bar"
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5["/"].attrs["name"] = value_1
            assert h5["/"].attrs["name"] == value_1
            group = h5["/"]

            # Make sure changes to the attribute consistent
            group._h5_group._v_attrs["name"] = value_2
            assert group.attrs["name"] == value_2

    def test_group_properties(self):
        with open_h5file(H5_TEST_FILE, mode="w", auto_groups=True) as h5:
            h5.create_array("/group1/group2/array", np.arange(3))
            h5.create_array("/group1/array1", np.arange(3))

            assert h5["/group1"].name == "group1"

            child_names = h5["/group1"].children_names
            assert sorted(child_names) == sorted(["group2", "array1"])

            sub_names = h5["/group1"].subgroup_names
            assert sub_names == ["group2"]

            assert h5["/group1"].root.name == "/"
            assert h5["/group1/group2"].root.name == "/"

    def test_iter_groups(self):
        with open_h5file(H5_TEST_FILE, mode="w", auto_groups=True) as h5:
            h5.create_array("/group1/array", np.arange(3))
            h5.create_array("/group1/subgroup/deep_array", np.arange(3))
            group = h5["/group1"]
            assert set(n.name for n in group.iter_groups()) == set(
                ["subgroup"]
            )

    def test_mapping_interface_for_file(self):
        with open_h5file(H5_TEST_FILE, mode="w", auto_groups=True) as h5:
            array = h5.create_array("/array", np.arange(3))
            h5.create_array("/group/deep_array", np.arange(3))
            # `deep_array` isn't a direct descendent and isn't counted.
            assert len(h5) == 2
            assert "/group" in h5
            assert "/array" in h5
            np.testing.assert_allclose(h5["/array"], array)
            assert set(n.name for n in h5) == set(["array", "group"])

    def test_mapping_interface_for_group(self):
        with open_h5file(H5_TEST_FILE, mode="w", auto_groups=True) as h5:
            array = h5.create_array("/group1/array", np.arange(3))
            h5.create_array("/group1/subgroup/deep_array", np.arange(3))
            group = h5["/group1"]
            # `deep_array` isn't a direct descendent and isn't counted.
            assert len(group) == 2
            assert "subgroup" in group
            assert "array" in group
            np.testing.assert_allclose(group["array"], array)
            assert set(n.name for n in group) == set(["array", "subgroup"])

    def test_group_str_and_repr(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            group = h5["/"]

            assert str(group) == str(group._h5_group)
            assert repr(group) == repr(group._h5_group)

    def test_attribute_translation(self):
        value_1 = (1, 2, 3)
        value_1_array = np.array(value_1)
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            h5["/"].attrs["name"] = value_1
            assert isinstance(h5["/"].attrs["name"], np.ndarray)
            np.testing.assert_allclose(h5["/"].attrs["name"], value_1_array)

    def test_get_attribute(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            attrs = h5["/"].attrs
            attrs["name"] = "hello"
            assert attrs.get("name") == attrs["name"]

    def test_del_attribute(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            attrs = h5["/"].attrs
            attrs["name"] = "hello"
            del attrs["name"]
            assert "name" not in attrs

    def test_get_attribute_default(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            assert h5["/"].attrs.get("missing", "null") == "null"

    def test_attribute_update(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            attrs = h5["/"].attrs
            attrs.update({"a": 1, "b": 2})
            assert attrs["a"] == 1
            assert attrs["b"] == 2
            attrs.update({"b": 20, "c": 30})
            assert attrs["b"] == 20
            assert attrs["c"] == 30

    def test_attribute_iteration_methods(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            attrs = h5["/"].attrs
            attrs["organ"] = "gallbladder"
            attrs["count"] = 42
            attrs["alpha"] = 0xFF

            items = list(attrs.items())

            assert all(isinstance(x, tuple) for x in items)

            # unfold the pairs
            keys, vals = [list(item) for item in zip(*items)]

            assert keys == list(attrs.keys())
            assert vals == list(attrs.values())

            # Check that __iter__ is consistent
            assert keys == list(iter(attrs))
            assert len(attrs) == len(keys)

    def test_bad_node_name(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            with self.assertRaises(ValueError):
                h5.create_array("/attrs", np.zeros(3))

    def test_bad_group_name(self):
        with open_h5file(H5_TEST_FILE, mode="w") as h5:
            with self.assertRaises(ValueError):
                h5.create_array("/attrs/array", np.zeros(3))

    def test_create_dict_with_H5File(self):
        data = {"a": 1}
        with temp_h5_file() as h5:
            h5.create_dict("/dict", data)
            assert isinstance(h5["/dict"], H5DictNode)
            assert h5["/dict"]["a"] == 1

    def test_create_dict_with_H5Group(self):
        node_path = "/bananas/dict"
        data = {"a": 1}
        with temp_h5_file() as h5:
            group = h5.create_group("/bananas")
            group.create_dict("dict", data)
            assert isinstance(h5[node_path], H5DictNode)
            assert h5[node_path]["a"] == 1

    def test_create_table_with_H5File(self):
        description = [("foo", "int"), ("bar", "float")]
        with temp_h5_file() as h5:
            h5.create_table("/table", description)
            tab = h5["/table"]
            assert isinstance(tab, H5TableNode)
            tab.append({"foo": (1,), "bar": (np.pi,)})
            assert tab.ix[0][0] == 1
            assert tab.ix[0][1] == np.pi

            h5.remove_node("/table")
            assert "/table" not in h5

    def test_create_table_with_H5Group(self):
        node_path = "/rhinocerous/table"
        description = [("foo", "int"), ("bar", "float")]
        with temp_h5_file() as h5:
            group = h5.create_group("/rhinocerous")
            group.create_table("table", description)
            tab = h5[node_path]
            assert isinstance(tab, H5TableNode)
            tab.append({"foo": (1,), "bar": (np.pi,)})
            assert tab.ix[0][0] == 1
            assert tab.ix[0][1] == np.pi

            group.remove_node("table")
            assert node_path not in h5



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests naming operations on PyFS contexts. """


# Standard library imports.
import os
import shutil
import unittest

# Enthought library imports.
from apptools.io import File
from apptools.naming.api import (
    DirContext,
    InvalidNameError,
    NameAlreadyBoundError,
    NameNotFoundError,
    NotContextError,
    PyFSContext,
)


class PyFSContextTestCase(unittest.TestCase):
    """ Tests naming operations on PyFS contexts. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        if os.path.exists("data"):
            shutil.rmtree("data", ignore_errors=True)
        if os.path.exists("other"):
            shutil.rmtree("other", ignore_errors=True)

        os.mkdir("data")
        os.mkdir("other")

        self.context = PyFSContext(path="data")
        self.context.create_subcontext("sub")
        self.context.bind("x", 123)
        self.context.bind("y", 321)

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        self.context = None

        shutil.rmtree("data")
        shutil.rmtree("other")

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_initialization(self):
        """ initialization of an existing context """

        context = PyFSContext(path="data")
        self.assertEqual(len(context.list_bindings("")), 3)

    def test_initialization_with_empty_environment(self):
        """ initialization with empty environmentt """

        context = PyFSContext(path="other", environment={})
        self.assertEqual(len(context.list_names("")), 0)

    def test_bind(self):
        """ pyfs context bind """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.bind, "", 1)

        # Bind a local file object.
        f = File(os.path.join(sub.path, "foo.py"))
        # f.create_file('print("foo!")\n')

        context.bind("sub/foo.py", f)
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Bind a reference to a non-local file.
        f = File("/tmp")
        context.bind("sub/tmp", f)
        self.assertEqual(len(sub.list_bindings("")), 2)
        self.assertEqual(context.lookup("sub/tmp").path, f.path)

        # Bind a reference to a non-local context.
        f = PyFSContext(path="other")
        context.bind("sub/other", f)
        self.assertEqual(len(sub.list_bindings("")), 3)
        self.assertIn(f.path, context.lookup("sub/other").path)

        # Bind a Python object.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 4)

        # Try to bind it again.
        self.assertRaises(NameAlreadyBoundError, context.bind, "sub/a", 1)

    def test_rebind(self):
        """ pyfs context rebind """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.rebind, "", 1)

        # Bind a name.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Rebind it.
        context.rebind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)

    def test_unbind(self):
        """ pyfs context unbind """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.unbind, "")

        # Bind a name.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Unbind it.
        context.unbind("sub/a")
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Try to unbind a non-existent name.
        self.assertRaises(NameNotFoundError, context.unbind, "sub/b")

    def test_rename(self):
        """ multi-context rename """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.rename, "", "x")
        self.assertRaises(InvalidNameError, context.rename, "x", "")

        # Bind a name.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Rename it.
        context.rename("sub/a", "sub/b")
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Lookup using the new name.
        self.assertEqual(context.lookup("sub/b"), 1)

        # Lookup using the old name.
        self.assertRaises(NameNotFoundError, context.lookup, "sub/a")

    def test_lookup(self):
        """ pyfs context lookup """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Bind a file object.
        f = File(os.path.join(sub.path, "foo.py"))
        # f.create_file('print("foo!")\n')

        context.bind("sub/foo.py", f)
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Look it up.
        self.assertEqual(context.lookup("sub/foo.py").path, f.path)

        # Bind a Python object.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 2)

        # Look it up.
        self.assertEqual(context.lookup("sub/a"), 1)

        # Looking up the Empty name returns the context itself.
        self.assertEqual(context.lookup(""), context)

        # Non-existent name.
        self.assertRaises(NameNotFoundError, context.lookup, "sub/b")

    def test_create_subcontext(self):
        """ pyfs context create sub-context """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.create_subcontext, "")

        # Create a sub-context.
        context.create_subcontext("sub/a")
        self.assertEqual(len(sub.list_bindings("")), 1)
        self.assertTrue(os.path.isdir(os.path.join(sub.path, "a")))

        # Try to bind it again.
        self.assertRaises(
            NameAlreadyBoundError, context.create_subcontext, "sub/a"
        )

    def test_destroy_subcontext(self):
        """ single context destroy sub-context """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")

        # Make sure that the sub-context is empty.
        self.assertEqual(len(sub.list_bindings("")), 0)

        # Empty name.
        self.assertRaises(InvalidNameError, context.destroy_subcontext, "")

        # Create a sub-context.
        context.create_subcontext("sub/a")
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Destroy it.
        context.destroy_subcontext("sub/a")
        self.assertEqual(len(sub.list_bindings("")), 0)
        self.assertTrue(not os.path.isdir(os.path.join(sub.path, "a")))

        # Bind a name.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)

        # Try to destroy it.
        self.assertRaises(NotContextError, context.destroy_subcontext, "sub/a")

        # Try to destroy a non-existent name.
        self.assertRaises(
            NameNotFoundError, context.destroy_subcontext, "sub/b"
        )

    def test_get_attributes(self):
        """ get attributes """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")
        self.assertIsInstance(sub, DirContext)

        #### Generic name resolution tests ####

        # Non-existent name.
        self.assertRaises(NameNotFoundError, context.get_attributes, "xx")

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.get_attributes, "xx/a")

        # Attempt to resolve via an existing name that is not a context.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)
        self.assertRaises(NotContextError, context.get_attributes, "sub/a/x")

        #### Operation specific tests ####

        # Attributes of the root context.
        attributes = context.get_attributes("")
        self.assertEqual(len(attributes), 0)

        # Attributes of a sub-context.
        attributes = context.get_attributes("sub")
        self.assertEqual(len(attributes), 0)

    def test_set_get_attributes(self):
        """ get and set attributes """

        defaults = {"colour": "blue"}

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")
        self.assertIsInstance(sub, DirContext)

        #### Generic name resolution tests ####

        # Non-existent name.
        self.assertRaises(
            NameNotFoundError, context.set_attributes, "xx", defaults
        )

        # Attempt to resolve via a non-existent context.
        self.assertRaises(
            NameNotFoundError, context.set_attributes, "xx/a", defaults
        )

        # Attempt to resolve via an existing name that is not a context.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)
        self.assertRaises(
            NotContextError, context.set_attributes, "sub/a/xx", defaults
        )

        #### Operation specific tests ####

        # Attributes of the root context.
        attributes = self.context.get_attributes("")
        self.assertEqual(len(attributes), 0)

        # Set the attributes.
        context.set_attributes("", defaults)
        attributes = context.get_attributes("")
        self.assertEqual(len(attributes), 1)
        self.assertEqual(attributes["colour"], "blue")

        # Attributes of a sub-context.
        attributes = context.get_attributes("sub")
        self.assertEqual(len(attributes), 0)

        # Set the attributes.
        context.set_attributes("sub", defaults)
        attributes = context.get_attributes("sub")
        self.assertEqual(len(attributes), 1)
        self.assertEqual(attributes["colour"], "blue")

    def test_namespace_name(self):
        """ get the name of a context within its namespace. """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")
        self.assertIsInstance(sub, DirContext)

        self.assertEqual(context.namespace_name, "data")
        self.assertEqual(sub.namespace_name, "data/sub")



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests for Python-esque '.ini' files. """


# Standard library imports.
import os
import tempfile
import unittest
from os.path import join

# Major package imports.
from importlib_resources import files

# Enthought library imports.
from .py_config_file import PyConfigFile


# This module's package.
PKG = "apptools.preferences.tests"


class PyConfigFileTestCase(unittest.TestCase):
    """ Tests for Python-esque '.ini' files. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        # The filenames of the example preferences files.
        self.example = os.fspath(files(PKG) / "py_config_example.ini")
        self.example_2 = os.fspath(files(PKG) / "py_config_example_2.ini")

    def tearDown(self):
        """ Called immediately after each test method has been called. """

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_load_from_filename(self):
        """ load from filename """

        config = PyConfigFile(self.example)

        self.assertEqual("blue", config["acme.ui"]["bgcolor"])
        self.assertEqual(50, config["acme.ui"]["width"])
        self.assertEqual(1.0, config["acme.ui"]["ratio"])
        self.assertEqual(True, config["acme.ui"]["visible"])
        self.assertEqual({"a": 1, "b": 2}, config["acme.ui"]["foo"])
        self.assertEqual([1, 2, 3, 4], config["acme.ui"]["bar"])
        self.assertEqual((1, "a", 6, 4), config["acme.ui"]["baz"])
        self.assertEqual("red", config["acme.ui.other"]["fred"])
        self.assertEqual(100, config["acme.ui.other"]["wilma"])
        self.assertEqual(90, config["tds.foogle"]["joe"])
        self.assertEqual("meerkat", config["simples"]["animal"])

    def test_load_from_file(self):
        """ load from file """

        config = PyConfigFile(open(self.example))

        self.assertEqual("blue", config["acme.ui"]["bgcolor"])
        self.assertEqual(50, config["acme.ui"]["width"])
        self.assertEqual(1.0, config["acme.ui"]["ratio"])
        self.assertEqual(True, config["acme.ui"]["visible"])
        self.assertEqual({"a": 1, "b": 2}, config["acme.ui"]["foo"])
        self.assertEqual([1, 2, 3, 4], config["acme.ui"]["bar"])
        self.assertEqual((1, "a", 6, 4), config["acme.ui"]["baz"])
        self.assertEqual("red", config["acme.ui.other"]["fred"])
        self.assertEqual(100, config["acme.ui.other"]["wilma"])
        self.assertEqual(90, config["tds.foogle"]["joe"])
        self.assertEqual("meerkat", config["simples"]["animal"])

    def test_save(self):
        """ save """

        config = PyConfigFile(open(self.example))

        self.assertEqual("blue", config["acme.ui"]["bgcolor"])
        self.assertEqual(50, config["acme.ui"]["width"])
        self.assertEqual(1.0, config["acme.ui"]["ratio"])
        self.assertEqual(True, config["acme.ui"]["visible"])
        self.assertEqual({"a": 1, "b": 2}, config["acme.ui"]["foo"])
        self.assertEqual([1, 2, 3, 4], config["acme.ui"]["bar"])
        self.assertEqual("red", config["acme.ui.other"]["fred"])
        self.assertEqual(100, config["acme.ui.other"]["wilma"])
        self.assertEqual(90, config["tds.foogle"]["joe"])
        self.assertEqual("meerkat", config["simples"]["animal"])

        # Save the config to another file.
        tmpdir = tempfile.mkdtemp()
        tmp = join(tmpdir, "tmp.ini")

        config.save(tmp)
        try:
            self.assertTrue(os.path.exists(tmp))

            # Make sure we can read the file back in and that we get the same
            # values!
            config = PyConfigFile(open(tmp))

            self.assertEqual("blue", config["acme.ui"]["bgcolor"])
            self.assertEqual(50, config["acme.ui"]["width"])
            self.assertEqual(1.0, config["acme.ui"]["ratio"])
            self.assertEqual(True, config["acme.ui"]["visible"])
            self.assertEqual({"a": 1, "b": 2}, config["acme.ui"]["foo"])
            self.assertEqual([1, 2, 3, 4], config["acme.ui"]["bar"])
            self.assertEqual((1, "a", 6, 4), config["acme.ui"]["baz"])
            self.assertEqual("red", config["acme.ui.other"]["fred"])
            self.assertEqual(100, config["acme.ui.other"]["wilma"])
            self.assertEqual(90, config["tds.foogle"]["joe"])
            self.assertEqual("meerkat", config["simples"]["animal"])

        finally:
            # Clean up!
            os.remove(tmp)
            os.rmdir(tmpdir)

    def test_load_multiple_files(self):
        """ load multiple files """

        config = PyConfigFile(self.example)

        self.assertEqual("blue", config["acme.ui"]["bgcolor"])
        self.assertEqual(50, config["acme.ui"]["width"])
        self.assertEqual(1.0, config["acme.ui"]["ratio"])
        self.assertEqual(True, config["acme.ui"]["visible"])
        self.assertEqual({"a": 1, "b": 2}, config["acme.ui"]["foo"])
        self.assertEqual([1, 2, 3, 4], config["acme.ui"]["bar"])
        self.assertEqual((1, "a", 6, 4), config["acme.ui"]["baz"])
        self.assertEqual("red", config["acme.ui.other"]["fred"])
        self.assertEqual(100, config["acme.ui.other"]["wilma"])
        self.assertEqual(90, config["tds.foogle"]["joe"])
        self.assertEqual("meerkat", config["simples"]["animal"])

        # Load another file.
        config.load(self.example_2)

        # Make sure we still have the unchanged values...
        self.assertEqual("red", config["acme.ui"]["bgcolor"])
        self.assertEqual(50, config["acme.ui"]["width"])
        self.assertEqual(1.0, config["acme.ui"]["ratio"])
        self.assertEqual(True, config["acme.ui"]["visible"])
        self.assertEqual({"a": 1, "b": 2}, config["acme.ui"]["foo"])
        self.assertEqual([1, 2, 3, 4], config["acme.ui"]["bar"])
        self.assertEqual((1, "a", 6, 4), config["acme.ui"]["baz"])
        self.assertEqual("red", config["acme.ui.other"]["fred"])
        self.assertEqual(100, config["acme.ui.other"]["wilma"])
        self.assertEqual(90, config["tds.foogle"]["joe"])
        self.assertEqual("meerkat", config["simples"]["animal"])

        # ... and the values that were overwritten...
        self.assertEqual("red", config["acme.ui"]["bgcolor"])

        # ... and that we have the new ones.
        self.assertEqual(42, config["acme.ui"]["bazzle"])

        # ... and that the new ones can refer to the old ones!
        self.assertEqual(180, config["acme.ui"]["blimey"])



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# Standard library imports
from datetime import datetime
import logging

# Enthought library imports.
from pyface.api import ImageResource, clipboard
from pyface.workbench.api import TraitsUIView
from traits.api import (
    Button,
    Instance,
    List,
    Property,
    Str,
    cached_property,
    on_trait_change,
)
from traitsui.api import View, Group, Item, CodeEditor, TabularEditor, spring
from traitsui.tabular_adapter import TabularAdapter

# Local imports
from apptools.logger.agent.quality_agent_view import QualityAgentView
from apptools.logger.plugin.logger_service import LoggerService

# Constants
_IMAGE_MAP = {
    logging.DEBUG: ImageResource("debug"),
    logging.INFO: ImageResource("info"),
    logging.WARNING: ImageResource("warning"),
    logging.ERROR: ImageResource("error"),
    logging.CRITICAL: ImageResource("crit_error"),
}


class LogRecordAdapter(TabularAdapter):
    """A TabularEditor adapter for logging.LogRecord objects."""

    columns = [
        ("Level", "level"),
        ("Date", "date"),
        ("Time", "time"),
        ("Message", "message"),
    ]
    column_widths = [80, 100, 120, -1]

    level_image = Property
    level_text = Property(Str)
    date_text = Property(Str)
    time_text = Property(Str)
    message_text = Property(Str)

    def get_width(self, object, trait, column):
        return self.column_widths[column]

    def _get_level_image(self):
        return _IMAGE_MAP[self.item.levelno]

    def _get_level_text(self):
        return self.item.levelname.capitalize()

    def _get_date_text(self):
        dt = datetime.fromtimestamp(self.item.created)
        return dt.date().isoformat()

    def _get_time_text(self):
        dt = datetime.fromtimestamp(self.item.created)
        return dt.time().isoformat()

    def _get_message_text(self):
        # Just display the first line of multiline messages, like stacktraces.
        msg = self.item.getMessage()
        msgs = msg.strip().split("\n")
        if len(msgs) > 1:
            suffix = "... [double click for details]"
        else:
            suffix = ""
        abbrev_msg = msgs[0] + suffix
        return abbrev_msg


class LoggerView(TraitsUIView):
    """The Workbench View showing the list of log items."""

    id = Str("apptools.logger.plugin.view.logger_view.LoggerView")
    name = Str("Logger")
    service = Instance(LoggerService)

    log_records = List(Instance(logging.LogRecord))
    formatted_records = Property(Str, depends_on="log_records")

    activated = Instance(logging.LogRecord)
    activated_text = Property(Str, depends_on="activated")
    reset_button = Button("Reset Logs")
    show_button = Button("Complete Text Log")
    copy_button = Button("Copy Log to Clipboard")

    code_editor = CodeEditor(lexer="null", show_line_numbers=False)
    log_records_editor = TabularEditor(
        adapter=LogRecordAdapter(), editable=False, activated="activated"
    )
    trait_view = View(
        Group(
            Item("log_records", editor=log_records_editor),
            Group(
                Item("reset_button"),
                spring,
                Item("show_button"),
                Item("copy_button"),
                orientation="horizontal",
                show_labels=False,
            ),
            show_labels=False,
        )
    )

    ###########################################################################
    # LogQueueHandler view interface
    ###########################################################################

    def update(self, force=False):
        """Update 'log_records' if our handler has new records or 'force' is
        set.
        """
        service = self.service
        if service.handler.has_new_records() or force:
            log_records = [
                rec
                for rec in service.handler.get()
                if rec.levelno >= service.preferences.level_
            ]
            log_records.reverse()
            self.log_records = log_records

    ###########################################################################
    # Private interface
    ###########################################################################

    @on_trait_change("service.preferences.level_")
    def _update_log_records(self):
        self.service.handler._view = self
        self.update(force=True)

    def _reset_button_fired(self):
        self.service.handler.reset()
        self.log_records = []

    def _show_button_fired(self):
        self.edit_traits(
            view=View(
                Item(
                    "formatted_records",
                    editor=self.code_editor,
                    style="readonly",
                    show_label=False,
                ),
                width=800,
                height=600,
                resizable=True,
                buttons=["OK"],
                title="Complete Text Log",
            )
        )

    def _copy_button_fired(self):
        clipboard.text_data = self.formatted_records

    @cached_property
    def _get_formatted_records(self):
        return "\n".join(
            [
                self.service.handler.formatter.format(record)
                for record in self.log_records
            ]
        )

    def _activated_changed(self):
        if self.activated is None:
            return
        msg = self.activated.getMessage()
        if self.service.preferences.enable_agent:
            dialog = QualityAgentView(msg=msg, service=self.service)
            dialog.open()
        else:
            self.edit_traits(
                view=View(
                    Item(
                        "activated_text",
                        editor=self.code_editor,
                        style="readonly",
                        show_label=False,
                    ),
                    width=800,
                    height=600,
                    resizable=True,
                    buttons=["OK"],
                    title="Log Message Detail",
                )
            )

    @cached_property
    def _get_activated_text(self):
        if self.activated is None:
            return ""
        else:
            return self.activated.getMessage()



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from ..type_registry import LazyRegistry

from .dummies import A, B, C, D, Mixed, Abstract, Concrete, ConcreteSubclass


class TestLazyRegistry(unittest.TestCase):
    def setUp(self):
        self.registry = LazyRegistry()
        # Bodge in a different importer.
        self.registry._import_object = "Imported {0}".format

    def test_deferred_push(self):
        self.registry.push("dummies:A", "foo:A")
        self.registry.push("dummies:C", "foo:C")
        self.assertEqual(self.registry.lookup_by_type(A), "Imported foo:A")
        self.assertEqual(self.registry.lookup_by_type(B), "Imported foo:A")
        self.assertEqual(self.registry.lookup_by_type(C), "Imported foo:C")
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_greedy_push(self):
        self.registry.push(A, "foo:A")
        self.registry.push(C, "foo:C")
        self.assertEqual(self.registry.lookup_by_type(A), "Imported foo:A")
        self.assertEqual(self.registry.lookup_by_type(B), "Imported foo:A")
        self.assertEqual(self.registry.lookup_by_type(C), "Imported foo:C")
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_mro(self):
        self.registry.push("dummies:A", "foo:A")
        self.registry.push("dummies:D", "foo:D")
        self.assertEqual(self.registry.lookup_by_type(Mixed), "Imported foo:A")

    def test_lookup_instance(self):
        self.registry.push(A, "foo:A")
        self.registry.push(C, "foo:C")
        self.assertEqual(self.registry.lookup(A()), "Imported foo:A")
        self.assertEqual(self.registry.lookup(B()), "Imported foo:A")
        self.assertEqual(self.registry.lookup(C()), "Imported foo:C")
        self.assertRaises(KeyError, self.registry.lookup, D())

    def test_abc(self):
        self.registry.push_abc(Abstract, "foo:Abstract")
        self.assertEqual(
            self.registry.lookup_by_type(Concrete), "Imported foo:Abstract"
        )
        self.assertEqual(
            self.registry.lookup_by_type(ConcreteSubclass),
            "Imported foo:Abstract",
        )



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""Tests for the file_path module.

"""

# Standard library imports.
import unittest
import os
import sys
from os.path import abspath, dirname, basename, join
from io import BytesIO

# 3rd party imports.
from importlib_resources import files

from apptools._testing.optional_dependencies import (
    numpy as np,
    requires_numpy,
)

# Enthought library imports.
if np is not None:
    from apptools.persistence import state_pickler
    from apptools.persistence import file_path


@requires_numpy
class Test:
    def __init__(self):
        self.f = file_path.FilePath()


@requires_numpy
class TestFilePath(unittest.TestCase):
    def setUp(self):
        # If the cwd is somewhere under /tmp, that confuses the tests below.
        # Use the directory containing this file, instead.
        test_cwd = os.fspath(files("apptools.persistence") / "")
        self.old_cwd = os.getcwd()
        os.chdir(test_cwd)

    def tearDown(self):
        os.chdir(self.old_cwd)

    def test_relative(self):
        """Test if relative paths are set correctly."""
        fname = "t.vtk"
        f = file_path.FilePath(fname)
        cwd = os.getcwd()
        # Trivial case of both in same dir.
        f.set_relative(abspath(join(cwd, "t.mv2")))
        self.assertEqual(f.rel_pth, fname)
        # Move one directory deeper.
        f.set_relative(abspath(join(cwd, "tests", "t.mv2")))
        self.assertEqual(f.rel_pth, join(os.pardir, fname))
        # Move one directory shallower.
        f.set_relative(abspath(join(dirname(cwd), "t.mv2")))
        diff = basename(cwd)
        self.assertEqual(f.rel_pth, join(diff, fname))
        # Test where the path is relative to the root.
        f.set(abspath(join("data", fname)))
        f.set_relative("/tmp/test.mv2")
        if sys.platform.startswith("win"):
            expect = os.pardir + abspath(join("data", fname))[2:]
        else:
            expect = os.pardir + abspath(join("data", fname))
        self.assertEqual(f.rel_pth, expect)

    def test_absolute(self):
        """Test if absolute paths are set corectly."""
        fname = "t.vtk"
        f = file_path.FilePath(fname)
        cwd = os.getcwd()
        # Easy case of both in same dir.
        f.set_absolute(join(cwd, "foo", "test", "t.mv2"))
        self.assertEqual(f.abs_pth, join(cwd, "foo", "test", fname))
        # One level lower.
        fname = join(os.pardir, "t.vtk")
        f.set(fname)
        f.set_absolute(join(cwd, "foo", "test", "t.mv2"))
        self.assertEqual(f.abs_pth, abspath(join(cwd, "foo", "test", fname)))
        # One level higher.
        fname = join("test", "t.vtk")
        f.set(fname)
        f.set_absolute(join(cwd, "foo", "t.mv2"))
        self.assertEqual(f.abs_pth, abspath(join(cwd, "foo", fname)))

    def test_pickle(self):
        """Test if pickler works correctly with FilePaths."""
        t = Test()
        t.f.set("t.vtk")
        cwd = os.getcwd()
        curdir = basename(cwd)

        # Create a dummy file in the parent dir.
        s = BytesIO()
        # Spoof its location.
        s.name = abspath(join(cwd, os.pardir, "t.mv2"))
        # Dump into it
        state_pickler.dump(t, s)

        # Rewind the stream
        s.seek(0)
        # "Move" the file elsewhere
        s.name = join(cwd, "foo", "test", "t.mv2")
        state = state_pickler.load_state(s)
        self.assertEqual(
            state.f.abs_pth, join(cwd, "foo", "test", curdir, "t.vtk")
        )

        # Create a dummy file in a subdir.
        s = BytesIO()
        # Spoof its location.
        s.name = abspath(join(cwd, "data", "t.mv2"))
        # Dump into it.
        state_pickler.dump(t, s)

        # Rewind the stream
        s.seek(0)
        # "Move" the file elsewhere
        s.name = join(cwd, "foo", "test", "t.mv2")
        state = state_pickler.load_state(s)
        self.assertEqual(state.f.abs_pth, join(cwd, "foo", "t.vtk"))



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""Tests for the version registry.

"""

# Standard library imports.
from importlib import reload
import unittest

# Enthought library imports.
from traits.api import HasTraits

from apptools._testing.optional_dependencies import (
    numpy as np,
    requires_numpy,
)
if np is not None:
    from apptools.persistence import version_registry, state_pickler


class Classic:
    __version__ = 0


class New(object):
    __version__ = 0


class TraitClass(HasTraits):
    __version__ = 0


class Test(New):
    __version__ = 1

    def __init__(self):
        self.a = Classic()


class Handler:
    def __init__(self):
        self.calls = []

    def upgrade(self, state, version):
        self.calls.append(("upgrade", state, version))

    def upgrade1(self, state, version):
        self.calls.append(("upgrade1", state, version))


@requires_numpy
class TestVersionRegistry(unittest.TestCase):
    def test_get_version(self):
        """Test the get_version function."""
        extra = [(("object", "builtins"), -1)]
        c = Classic()
        v = version_registry.get_version(c)
        res = extra + [(("Classic", __name__), 0)]
        self.assertEqual(v, res)
        state = state_pickler.get_state(c)
        self.assertEqual(state.__metadata__["version"], res)

        n = New()
        v = version_registry.get_version(n)
        res = extra + [(("New", __name__), 0)]
        self.assertEqual(v, res)
        state = state_pickler.get_state(n)
        self.assertEqual(state.__metadata__["version"], res)

        t = TraitClass()
        v = version_registry.get_version(t)
        res = extra + [
            (("CHasTraits", "traits.ctraits"), -1),
            (("HasTraits", "traits.has_traits"), -1),
            (("TraitClass", __name__), 0),
        ]
        self.assertEqual(v, res)
        state = state_pickler.get_state(t)
        self.assertEqual(state.__metadata__["version"], res)

    def test_reload(self):
        """Test if the registry is reload safe."""
        # A dummy handler.
        def h(x, y):
            pass

        registry = version_registry.registry
        registry.register("A", __name__, h)
        self.assertEqual(registry.handlers.get(("A", __name__)), h)
        reload(version_registry)
        registry = version_registry.registry
        self.assertEqual(registry.handlers.get(("A", __name__)), h)
        del registry.handlers[("A", __name__)]
        self.assertNotIn(("A", __name__), registry.handlers)

    def test_update(self):
        """Test if update method calls the handlers in order."""
        registry = version_registry.registry

        # First an elementary test.
        c = Classic()
        state = state_pickler.get_state(c)
        h = Handler()
        registry.register("Classic", __name__, h.upgrade)
        c1 = state_pickler.create_instance(state)
        state_pickler.set_state(c1, state)
        self.assertEqual(h.calls, [("upgrade", state, 0)])
        # Remove the handler.
        registry.unregister("Classic", __name__)

        # Now check to see if this works for inheritance trees.
        t = Test()
        state = state_pickler.get_state(t)
        h = Handler()
        registry.register("Classic", __name__, h.upgrade)
        registry.register("New", __name__, h.upgrade)
        registry.register("Test", __name__, h.upgrade1)
        t1 = state_pickler.create_instance(state)
        state_pickler.set_state(t1, state)
        # This should call New handler, then the Test and then
        # Classic.
        self.assertEqual(
            h.calls,
            [
                ("upgrade", state, 0),
                ("upgrade1", state, 1),
                ("upgrade", state.a, 0),
            ],
        )



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Tests the default directory context. """


# Enthought library imports.
from apptools.naming.api import (
    DirContext,
    NameNotFoundError,
    NotContextError,
)

# Local imports.
from .test_context import ContextTestCase


class DirContextTestCase(ContextTestCase):
    """ Tests the default directory context. """

    ###########################################################################
    # 'ContextTestCase' interface.
    ###########################################################################

    def create_context(self):
        """ Creates the context that we are testing. """

        return DirContext()

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_get_attributes(self):
        """ get attributes """

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")
        self.assertIsInstance(sub, DirContext)

        #### Generic name resolution tests ####

        # Non-existent name.
        self.assertRaises(NameNotFoundError, context.get_attributes, "x")

        # Attempt to resolve via a non-existent context.
        self.assertRaises(NameNotFoundError, context.get_attributes, "x/a")

        # Attempt to resolve via an existing name that is not a context.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)
        self.assertRaises(NotContextError, context.get_attributes, "sub/a/x")

        #### Operation specific tests ####

        # Attributes of the root context.
        attributes = self.context.get_attributes("")
        self.assertEqual(len(attributes), 0)

        # Attributes of a sub-context.
        attributes = context.get_attributes("sub")
        self.assertEqual(len(attributes), 0)

    def test_set_get_attributes(self):
        """ get and set attributes """

        defaults = {"colour": "blue"}

        # Convenience.
        context = self.context
        sub = self.context.lookup("sub")
        self.assertIsInstance(sub, DirContext)

        #### Generic name resolution tests ####

        # Non-existent name.
        self.assertRaises(
            NameNotFoundError, context.set_attributes, "x", defaults
        )

        # Attempt to resolve via a non-existent context.
        self.assertRaises(
            NameNotFoundError, context.set_attributes, "x/a", defaults
        )

        # Attempt to resolve via an existing name that is not a context.
        context.bind("sub/a", 1)
        self.assertEqual(len(sub.list_bindings("")), 1)
        self.assertRaises(
            NotContextError, context.set_attributes, "sub/a/xx", defaults
        )

        #### Operation specific tests ####

        # Attributes of the root context.
        attributes = self.context.get_attributes("")
        self.assertEqual(len(attributes), 0)

        # Set the attributes.
        context.set_attributes("", defaults)
        attributes = context.get_attributes("")
        self.assertEqual(len(attributes), 1)
        self.assertEqual(attributes["colour"], "blue")

        # Attributes of a sub-context.
        attributes = context.get_attributes("sub")
        self.assertEqual(len(attributes), 0)

        # Set the attributes.
        context.set_attributes("sub", defaults)
        attributes = context.get_attributes("sub")
        self.assertEqual(len(attributes), 1)
        self.assertEqual(attributes["colour"], "blue")
